@@ -14,12 +14,14 @@
 #    under the License.
 
 import collections
-import imp
+from importlib import util as imp_util
 import os
 
+from keystoneauth1 import loading as ks_loading
 from neutron_lib.api import extensions as api_extensions
 from neutron_lib import exceptions
 from neutron_lib.plugins import directory
+from openstack import connection
 from oslo_config import cfg
 from oslo_log import log as logging
 from oslo_middleware import base
@@ -28,21 +30,21 @@ import webob.dec
 import webob.exc
 
 from neutron._i18n import _
+from neutron.api import wsgi
 from neutron import extensions as core_extensions
 from neutron.plugins.common import constants as const
 from neutron.services import provider_configuration
-from neutron import wsgi
-
 
 LOG = logging.getLogger(__name__)
 
 
 EXTENSION_SUPPORTED_CHECK_MAP = {}
 _PLUGIN_AGNOSTIC_EXTENSIONS = set()
+_NOVA_CONNECTION = None
 
 
 def register_custom_supported_check(alias, f, plugin_agnostic=False):
-    '''Register a custom function to determine if extension is supported.
+    """Register a custom function to determine if extension is supported.
 
     Consequent calls for the same alias replace the registered function.
 
@@ -51,7 +53,7 @@ def register_custom_supported_check(alias, f, plugin_agnostic=False):
     :param plugin_agnostic: if False, don't require a plugin to claim support
     with supported_extension_aliases. If True, a plugin must claim the
     extension is supported.
-    '''
+    """
 
     EXTENSION_SUPPORTED_CHECK_MAP[alias] = f
     if plugin_agnostic:
@@ -448,7 +450,9 @@ class ExtensionManager(object):
                 mod_name, file_ext = os.path.splitext(os.path.split(f)[-1])
                 ext_path = os.path.join(path, f)
                 if file_ext.lower() == '.py' and not mod_name.startswith('_'):
-                    mod = imp.load_source(mod_name, ext_path)
+                    spec = imp_util.spec_from_file_location(mod_name, ext_path)
+                    mod = imp_util.module_from_spec(spec)
+                    spec.loader.exec_module(mod)
                     ext_name = mod_name.capitalize()
                     new_ext_class = getattr(mod, ext_name, None)
                     if not new_ext_class:
@@ -645,3 +649,28 @@ def append_api_extensions_path(paths):
     paths = list(set([cfg.CONF.api_extensions_path] + paths))
     cfg.CONF.set_override('api_extensions_path',
                           ':'.join([p for p in paths if p]))
+
+
+class ProjectIdMiddleware(base.ConfigurableMiddleware):
+
+    @webob.dec.wsgify
+    def __call__(self, req):
+        # NOTE(ralonsoh): this method uses Nova Keystone user to retrieve the
+        # project because (1) it is allowed to retrieve the projects and (2)
+        # Neutron avoids adding another user section in the configuration
+        # (Nova user will be always used).
+        global _NOVA_CONNECTION
+        project = req.params.get('project_id') or req.params.get('tenant_id')
+        if project:
+            if not _NOVA_CONNECTION:
+                auth = ks_loading.load_auth_from_conf_options(cfg.CONF, 'nova')
+                keystone_session = ks_loading.load_session_from_conf_options(
+                    cfg.CONF, 'nova', auth=auth)
+                _NOVA_CONNECTION = connection.Connection(
+                    session=keystone_session, oslo_conf=cfg.CONF,
+                    connect_retries=cfg.CONF.http_retries)
+            if not _NOVA_CONNECTION.get_project(project):
+                return webob.exc.HTTPNotFound(
+                    comment='Project %s does not exist' % project)
+
+        return req.get_response(self.application)
