@@ -25,7 +25,6 @@ from neutron_lib.api.definitions import dhcpagentscheduler
 from neutron_lib.api.definitions import dns
 from neutron_lib.api.definitions import dns_domain_keywords
 from neutron_lib.api.definitions import dns_domain_ports
-from neutron_lib.api.definitions import empty_string_filtering
 from neutron_lib.api.definitions import expose_port_forwarding_in_fip
 from neutron_lib.api.definitions import external_net
 from neutron_lib.api.definitions import extra_dhcp_opt
@@ -59,6 +58,7 @@ from neutron_lib.api.definitions import port_device_profile
 from neutron_lib.api.definitions import port_hardware_offload_type
 from neutron_lib.api.definitions import port_mac_address_regenerate
 from neutron_lib.api.definitions import port_numa_affinity_policy
+from neutron_lib.api.definitions import port_numa_affinity_policy_socket
 from neutron_lib.api.definitions import port_resource_request
 from neutron_lib.api.definitions import port_security
 from neutron_lib.api.definitions import portbindings
@@ -85,8 +85,11 @@ from neutron_lib.api.definitions import segment as seg_def
 from neutron_lib.api.definitions import sorting
 from neutron_lib.api.definitions import stateful_security_group
 from neutron_lib.api.definitions import subnet_dns_publish_fixed_ip
+from neutron_lib.api.definitions import subnet_external_network
 from neutron_lib.api.definitions import subnet_service_types
 from neutron_lib.api.definitions import subnetpool_prefix_ops
+from neutron_lib.api.definitions import tag_creation
+from neutron_lib.api.definitions import tap_mirror
 from neutron_lib.api.definitions import trunk
 from neutron_lib.api.definitions import uplink_status_propagation
 from neutron_lib.api.definitions import vlantransparent
@@ -94,6 +97,7 @@ from neutron_lib.api.definitions import vpn
 from neutron_lib.api.definitions import vpn_endpoint_groups
 from neutron_lib import constants
 
+from neutron.extensions import port_trusted_vif
 from neutron.extensions import quotasv2_detail
 from neutron.extensions import security_groups_default_rules
 
@@ -141,7 +145,6 @@ ML2_SUPPORTED_API_EXTENSIONS = [
     default_subnetpools.ALIAS,
     dhcpagentscheduler.ALIAS,
     dns.ALIAS,
-    empty_string_filtering.ALIAS,
     external_net.ALIAS,
     extra_dhcp_opt.ALIAS,
     filter_validation.ALIAS,
@@ -155,7 +158,9 @@ ML2_SUPPORTED_API_EXTENSIONS = [
     port_hardware_offload_type.ALIAS,
     port_mac_address_regenerate.ALIAS,
     port_numa_affinity_policy.ALIAS,
+    port_numa_affinity_policy_socket.ALIAS,
     port_security.ALIAS,
+    port_trusted_vif.ALIAS,
     provider_net.ALIAS,
     port_resource_request.ALIAS,
     qos.ALIAS,
@@ -183,7 +188,9 @@ ML2_SUPPORTED_API_EXTENSIONS = [
     'standard-attr-tag',
     'standard-attr-timestamp',
     subnetpool_prefix_ops.ALIAS,
+    subnet_external_network.ALIAS,
     subnet_service_types.ALIAS,
+    tag_creation.ALIAS,
     trunk.ALIAS,
     seg_def.ALIAS,
     expose_port_forwarding_in_fip.ALIAS,
@@ -200,4 +207,5 @@ ML2_SUPPORTED_API_EXTENSIONS = [
     firewall_v2.ALIAS,
     firewall_v2_stdattrs.ALIAS,
     uplink_status_propagation.ALIAS,
+    tap_mirror.ALIAS,
 ]
