@@ -12,6 +12,7 @@
 
 import collections
 import copy
+import functools
 import inspect
 import os
 import random
@@ -27,6 +28,7 @@ from neutron_lib.api import validators
 from neutron_lib import constants as const
 from neutron_lib import context as n_context
 from neutron_lib import exceptions as n_exc
+from neutron_lib.plugins import constants as plugin_constants
 from neutron_lib.plugins import directory
 from neutron_lib.utils import net as n_utils
 from oslo_concurrency import processutils
@@ -35,13 +37,13 @@ from oslo_log import log
 from oslo_serialization import jsonutils
 from oslo_utils import netutils
 from oslo_utils import strutils
-from ovsdbapp.backend.ovs_idl import idlutils
+from ovsdbapp.backend.ovs_idl import rowview
 from ovsdbapp import constants as ovsdbapp_const
+from pecan import util as pecan_util
 import tenacity
 
 from neutron._i18n import _
 from neutron.common import _constants as n_const
-from neutron.common import ipv6_utils
 from neutron.common.ovn import constants
 from neutron.common.ovn import exceptions as ovn_exc
 from neutron.common import utils as common_utils
@@ -65,7 +67,7 @@ BPInfo = collections.namedtuple(
 
 HAChassisGroupInfo = collections.namedtuple(
     'HAChassisGroupInfo', ['group_name', 'chassis_list', 'az_hints',
-                           'ignore_chassis'])
+                           'ignore_chassis', 'external_ids'])
 
 
 class OvsdbClientCommand(object):
@@ -132,6 +134,68 @@ class OvsdbClientTransactCommand(OvsdbClientCommand):
     COMMAND = 'transact'
 
 
+def ovn_context(txn_var_name='txn', idl_var_name='idl'):
+    """Provide an OVN IDL transaction context
+
+    This decorator provides an OVN IDL database transaction context if the
+    'txn_var_name' variable, that should have an
+    ``ovsdbapp.backend.ovs_idl.transaction.Transaction`` derived object, is
+    empty. In that case (an empty transaction), that means the decorated method
+    has been called outside a transaction. In this case, the decorator creates
+    a transaction from the provided IDL and assigns it to the 'txn_var_name'
+    variable.
+    """
+    def decorator(f):
+        signature = inspect.signature(f)
+        if (txn_var_name not in signature.parameters or
+                idl_var_name not in signature.parameters):
+            msg = (_('Could not find variables %s and %s in the method '
+                     'signature') %
+                   (txn_var_name, idl_var_name))
+            raise RuntimeError(msg)
+
+        def retrieve_parameter(param_name, _args, _kwargs):
+            # Position of the parameter "param_name" in the "args" tuple.
+            param_index = pecan_util.getargspec(f).args.index(param_name)
+            try:  # Parameter passed as a positional argument.
+                value = _args[param_index]
+            except IndexError:  # Parameter passed as keyword argument.
+                # Reset the "param_index" value, that means the parameter is
+                # passed as keyword and if needed, it will be replaced in
+                # "kwargs".
+                param_index = None
+                try:
+                    value = _kwargs[param_name]
+                except KeyError:
+                    # Parameter is not passed as keyword nor positional, read
+                    # the keyword default value.
+                    value = signature.parameters[param_name].default
+
+            return value, param_index
+
+        @functools.wraps(f)
+        def wrapped(*args, **kwargs):
+            _txn, txn_index = retrieve_parameter(txn_var_name, args, kwargs)
+            _idl, idl_index = retrieve_parameter(idl_var_name, args, kwargs)
+            if not _txn and not _idl:
+                msg = (_('If no transaction is defined, it is needed at least '
+                         'an IDL connection'))
+                raise RuntimeError(msg)
+
+            if not _txn:
+                with _idl.transaction(check_error=True) as new_txn:
+                    if txn_index:
+                        args = (args[:txn_index] + (new_txn,) +
+                                args[txn_index + 1:])
+                    else:
+                        kwargs[txn_var_name] = new_txn
+                    return f(*args, **kwargs)
+            else:
+                return f(*args, **kwargs)
+        return wrapped
+    return decorator
+
+
 def ovn_name(id):
     # The name of the OVN entry will be neutron-<UUID>
     # This is due to the fact that the OVN application checks if the name
@@ -197,15 +261,6 @@ def ovn_pg_addrset_name(sg_id, ip_version):
     return ('pg-%s-%s' % (sg_id, ip_version)).replace('-', '_')
 
 
-def ovn_ag_addrset_name(ag_id, ip_version):
-    # The name of the address set for the given address group id and ip
-    # version. The format is:
-    #   ag-<address group uuid>-<ip version>
-    # with all '-' replaced with '_'. This replacement is necessary
-    # because OVN doesn't support '-' in an address set name.
-    return ('ag-%s-%s' % (ag_id, ip_version)).replace('-', '_')
-
-
 def ovn_port_group_name(sg_id):
     # The name of the port group for the given security group id.
     # The format is: pg-<security group uuid>.
@@ -402,9 +457,7 @@ def validate_and_get_data_from_binding_profile(port):
         if pbp_param_set.vnic_type:
             if pbp_param_set.vnic_type != vnic_type:
                 continue
-            if (capabilities and
-                    pbp_param_set.capability is not None and
-                    pbp_param_set.capability not in capabilities):
+            if capabilities and pbp_param_set.capability not in capabilities:
                 continue
         param_set = pbp_param_set.param_set
         param_keys = param_set.keys()
@@ -496,7 +549,6 @@ def get_revision_number(resource, resource_type):
                          constants.TYPE_ROUTERS,
                          constants.TYPE_ROUTER_PORTS,
                          constants.TYPE_SECURITY_GROUPS,
-                         constants.TYPE_ADDRESS_GROUPS,
                          constants.TYPE_FLOATINGIPS, constants.TYPE_SUBNETS):
         return resource['revision_number']
     else:
@@ -572,8 +624,13 @@ def sort_ips_by_version(addresses):
     return ip_map
 
 
-def is_lsp_router_port(port):
-    return port.get('device_owner') in const.ROUTER_PORT_OWNERS
+def is_lsp_router_port(neutron_port=None, lsp=None):
+    if neutron_port:
+        return neutron_port.get('device_owner') in const.ROUTER_PORT_OWNERS
+    elif lsp:
+        return (lsp.external_ids.get(constants.OVN_DEVICE_OWNER_EXT_ID_KEY) in
+                const.ROUTER_PORT_OWNERS)
+    return False
 
 
 def get_lrouter_ext_gw_static_route(ovn_router):
@@ -619,7 +676,7 @@ def get_system_dns_resolvers(resolver_file=DNS_RESOLVER_FILE):
                         netutils.is_valid_ipv6(line))
             if valid_ip:
                 if netutils.is_valid_ipv6(line):
-                    line = ipv6_utils.get_noscope_ipv6(line)
+                    line = netutils.get_noscope_ipv6(line)
                 resolvers.append(line)
 
     return resolvers
@@ -651,10 +708,12 @@ def get_port_subnet_ids(port):
     return [f['subnet_id'] for f in fixed_ips]
 
 
-def get_method_class(method):
-    if not inspect.ismethod(method):
+def get_method_class(method_or_class):
+    if not inspect.ismethod(method_or_class):
+        if inspect.isclass(method_or_class):
+            return method_or_class
         return
-    return method.__self__.__class__
+    return method_or_class.__self__.__class__
 
 
 def ovn_metadata_name(id_):
@@ -1000,44 +1059,6 @@ def get_subnets_address_scopes(context, subnets_by_id, fixed_ips, ml2_plugin):
     return address4_scope_id, address6_scope_id
 
 
-def _get_info_for_ha_chassis_group(context, port_id, network_id, sb_idl):
-    """Get the common required information to create a HA Chassis Group.
-
-    :param context:    Neutron API context
-    :param port_id:    The port ID
-    :param network_id: The network ID
-    :param sb_idl:     OVN SB IDL
-    :returns:          An instance of HAChassisGroupInfo
-    """
-    ignore_chassis = set()
-    # If there are Chassis marked for hosting external ports create a HA
-    # Chassis Group per external port, otherwise do it at the network level
-    chassis_list = sb_idl.get_extport_chassis_from_cms_options()
-    if chassis_list:
-        group_name = ovn_extport_chassis_group_name(port_id)
-        # Check if the port is bound to a chassis and if so, ignore that
-        # chassis when building the HA Chassis Group to ensure the
-        # external port is bound to a different chassis than the VM
-        ignore_chassis = sb_idl.get_chassis_host_for_port(port_id)
-        LOG.debug('HA Chassis Group %s is based on external port %s '
-                  '(network %s)', group_name, port_id, network_id)
-    else:
-        chassis_list = sb_idl.get_gateway_chassis_from_cms_options(
-            name_only=False)
-        group_name = ovn_name(network_id)
-        LOG.debug('HA Chassis Group %s is based on network %s',
-                  group_name, network_id)
-
-    # Get the Availability Zones hints
-    plugin = directory.get_plugin()
-    az_hints = common_utils.get_az_hints(
-        plugin.get_network(context, network_id))
-
-    return HAChassisGroupInfo(
-        group_name=group_name, chassis_list=chassis_list, az_hints=az_hints,
-        ignore_chassis=ignore_chassis)
-
-
 def _filter_candidates_for_ha_chassis_group(hcg_info):
     """Filter a list of chassis candidates for a given HA Chassis Group.
 
@@ -1065,40 +1086,39 @@ def _filter_candidates_for_ha_chassis_group(hcg_info):
     return candidates
 
 
-def sync_ha_chassis_group(context, port_id, network_id, nb_idl, sb_idl, txn):
+def _sync_ha_chassis_group(nb_idl, hcg_info, txn):
     """Return the UUID of the HA Chassis Group or the HA Chassis Group cmd.
 
-    Given the Neutron Network ID, this method will return (or create
-    and then return) the appropriate HA Chassis Group the external
-    port (in that network) needs to be associated with.
+    This method will return (or create and then return) the appropriate HA
+    Chassis Group for the resource specified in ``HAChassisGroupInfo``,
+    which can be generated from a network or a router.
 
-    :param context: Neutron API context
-    :param port_id: The port ID
-    :param network_id: The network ID
     :param nb_idl: OVN NB IDL
-    :param sb_idl: OVN SB IDL
+    :param hcg_info: HA Chassis Group information named tuple
+                     (``HAChassisGroupInfo``)
     :param txn: The ovsdbapp transaction object
-    :returns: The HA Chassis Group UUID or the HA Chassis Group command object
+    :returns: The HA Chassis Group UUID or the HA Chassis Group command object,
+              The name of the Chassis with the highest priority (could be None)
     """
     # If there are Chassis marked for hosting external ports create a HA
     # Chassis Group per external port, otherwise do it at the network level
-    hcg_info = _get_info_for_ha_chassis_group(context, port_id, network_id,
-                                              sb_idl)
     candidates = _filter_candidates_for_ha_chassis_group(hcg_info)
 
     # Try to get the HA Chassis Group or create if it doesn't exist
-    ha_ch_grp = ha_ch_grp_cmd = None
-    try:
-        ha_ch_grp = nb_idl.ha_chassis_group_get(
-            hcg_info.group_name).execute(check_error=True)
-    except idlutils.RowNotFound:
-        ext_ids = {constants.OVN_AZ_HINTS_EXT_ID_KEY: ','.join(
-            hcg_info.az_hints)}
-        ha_ch_grp_cmd = txn.add(nb_idl.ha_chassis_group_add(
-            hcg_info.group_name, may_exist=True, external_ids=ext_ids))
+    ha_ch_grp_cmd = txn.add(nb_idl.ha_chassis_group_add(
+        hcg_info.group_name, may_exist=True,
+        external_ids=hcg_info.external_ids))
+
+    if isinstance(ha_ch_grp_cmd.result, rowview.RowView):
+        # The HA chassis group existed before this transaction.
+        ha_ch_grp = ha_ch_grp_cmd.result
+    else:
+        # The HA chassis group is being created in this transaction.
+        ha_ch_grp = None
 
     max_chassis_number = constants.MAX_CHASSIS_IN_HA_GROUP
     priority = constants.HA_CHASSIS_GROUP_HIGHEST_PRIORITY
+    high_prio_ch_name = None
 
     # Check if the HA Chassis Group existed before. If so, re-calculate
     # the canditates in case something changed and keep the highest priority
@@ -1119,6 +1139,7 @@ def sync_ha_chassis_group(context, port_id, network_id, nb_idl, sb_idl, txn):
         if (high_prio_ch and
                 high_prio_ch.chassis_name in candidates):
             # If found, keep it as the highest priority chassis in the group
+            high_prio_ch_name = high_prio_ch.chassis_name
             txn.add(nb_idl.ha_chassis_group_add_chassis(
                 hcg_info.group_name, high_prio_ch.chassis_name,
                 priority=priority))
@@ -1139,11 +1160,68 @@ def sync_ha_chassis_group(context, port_id, network_id, nb_idl, sb_idl, txn):
         txn.add(nb_idl.ha_chassis_group_add_chassis(
             hcg_info.group_name, ch, priority=priority))
         priority -= 1
+        if not high_prio_ch_name:
+            high_prio_ch_name = ch
 
-    LOG.info('HA Chassis Group %s synchronized', hcg_info.group_name)
+    LOG.info('HA Chassis Group %s synchronized; highest priority chassis %s',
+             hcg_info.group_name, high_prio_ch_name)
     # Return the existing register UUID or the HA chassis group creation
     # command (see ovsdbapp ``HAChassisGroupAddChassisCommand`` class).
-    return ha_ch_grp.uuid if ha_ch_grp else ha_ch_grp_cmd
+    return ha_ch_grp.uuid if ha_ch_grp else ha_ch_grp_cmd, high_prio_ch_name
+
+
+@ovn_context(idl_var_name='nb_idl')
+def sync_ha_chassis_group_router(context, nb_idl, sb_idl, router_id, txn):
+    """Syncs the HA Chassis Group for a given router"""
+    chassis_list = sb_idl.get_gateway_chassis_from_cms_options(
+        name_only=False)
+    group_name = ovn_name(router_id)
+    LOG.debug('HA Chassis Group %s is based on router %s',
+              group_name, router_id)
+    plugin = directory.get_plugin(plugin_constants.L3)
+    resource = plugin.get_router(context, router_id,
+                                 fields=['availability_zone_hints'])
+    az_hints = common_utils.get_az_hints(resource)
+    external_ids = {constants.OVN_AZ_HINTS_EXT_ID_KEY: ','.join(az_hints),
+                    constants.OVN_ROUTER_ID_EXT_ID_KEY: router_id}
+    hcg_info = HAChassisGroupInfo(
+        group_name=group_name, chassis_list=chassis_list, az_hints=az_hints,
+        ignore_chassis=set(), external_ids=external_ids)
+    return _sync_ha_chassis_group(nb_idl, hcg_info, txn)
+
+
+@ovn_context(idl_var_name='nb_idl')
+def sync_ha_chassis_group_network(context, nb_idl, sb_idl, port_id,
+                                  network_id, txn):
+    """Syncs the HA Chassis Group for a given network"""
+    # If there are Chassis marked for hosting external ports create a HA
+    # Chassis Group per external port, otherwise do it at the network
+    # level
+    chassis_list = sb_idl.get_extport_chassis_from_cms_options()
+    if chassis_list:
+        group_name = ovn_extport_chassis_group_name(port_id)
+        # Check if the port is bound to a chassis and if so, ignore that
+        # chassis when building the HA Chassis Group to ensure the
+        # external port is bound to a different chassis than the VM
+        ignore_chassis = sb_idl.get_chassis_host_for_port(port_id)
+        LOG.debug('HA Chassis Group %s is based on external port %s '
+                  '(network %s)', group_name, port_id, network_id)
+    else:
+        chassis_list = sb_idl.get_gateway_chassis_from_cms_options(
+            name_only=False)
+        group_name = ovn_name(network_id)
+        ignore_chassis = set()
+        LOG.debug('HA Chassis Group %s is based on network %s',
+                  group_name, network_id)
+
+    plugin = directory.get_plugin()
+    resource = plugin.get_network(context, network_id)
+    az_hints = common_utils.get_az_hints(resource)
+    external_ids = {constants.OVN_AZ_HINTS_EXT_ID_KEY: ','.join(az_hints)}
+    hcg_info = HAChassisGroupInfo(
+        group_name=group_name, chassis_list=chassis_list, az_hints=az_hints,
+        ignore_chassis=ignore_chassis, external_ids=external_ids)
+    return _sync_ha_chassis_group(nb_idl, hcg_info, txn)
 
 
 def get_port_type_virtual_and_parents(subnets_by_id, fixed_ips, network_id,
