@@ -19,11 +19,13 @@
 """Utilities and helper functions."""
 
 import functools
+import hashlib
+import hmac
 import importlib
 import os
 import os.path
-import random
 import re
+import secrets
 import signal
 import socket
 import sys
@@ -44,11 +46,12 @@ from neutron_lib.utils import helpers
 from oslo_config import cfg
 from oslo_db import exception as db_exc
 from oslo_log import log as logging
+from oslo_utils import encodeutils
 from oslo_utils import excutils
 from oslo_utils import timeutils
 from oslo_utils import uuidutils
+from oslo_utils import versionutils
 from osprofiler import profiler
-import pkg_resources
 from sqlalchemy.dialects.mysql import dialect as mysql_dialect
 from sqlalchemy.dialects.postgresql import dialect as postgresql_dialect
 from sqlalchemy.dialects.sqlite import dialect as sqlite_dialect
@@ -57,6 +60,8 @@ from sqlalchemy.sql.expression import func as sql_func
 import neutron
 from neutron._i18n import _
 from neutron.api import api_common
+from neutron.conf import service as conf_service
+
 
 TIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
 LOG = logging.getLogger(__name__)
@@ -121,12 +126,12 @@ def _subprocess_setup():
     signal.signal(signal.SIGPIPE, signal.SIG_DFL)
 
 
-def subprocess_popen(args, stdin=None, stdout=None, stderr=None, shell=False,
-                     env=None, preexec_fn=_subprocess_setup, close_fds=True):
+def subprocess_popen(args, stdin=None, stdout=None, stderr=None, env=None,
+                     preexec_fn=_subprocess_setup, close_fds=True):
 
-    return subprocess.Popen(args, shell=shell, stdin=stdin, stdout=stdout,
-                            stderr=stderr, preexec_fn=preexec_fn,
-                            close_fds=close_fds, env=env)
+    return subprocess.Popen(args, stdin=stdin, stdout=stdout, stderr=stderr,
+                            preexec_fn=preexec_fn, close_fds=close_fds,
+                            env=env)
 
 
 def get_first_host_ip(net, ip_version):
@@ -188,9 +193,7 @@ def get_other_dvr_serviced_device_owners(host_dvr_for_dhcp=True):
     prefix, not a complete device_owner name, so should be handled
     separately (see is_dvr_serviced() below)
     """
-    device_owners = [n_const.DEVICE_OWNER_LOADBALANCER,
-                     n_const.DEVICE_OWNER_LOADBALANCERV2,
-                     trunk_constants.TRUNK_SUBPORT_OWNER]
+    device_owners = [trunk_constants.TRUNK_SUBPORT_OWNER]
     if host_dvr_for_dhcp:
         device_owners.append(n_const.DEVICE_OWNER_DHCP)
     return device_owners
@@ -205,10 +208,7 @@ def get_dvr_allowed_address_pair_device_owners():
     Later if other device owners are used for allowed_address_pairs those
     device_owners should be added to the list below.
     """
-    # TODO(Swami): Convert these methods to constants.
-    # Add the constants variable to the neutron-lib
-    return [n_const.DEVICE_OWNER_LOADBALANCER,
-            n_const.DEVICE_OWNER_LOADBALANCERV2]
+    return []
 
 
 def is_dvr_serviced(device_owner):
@@ -311,14 +311,33 @@ def cidr_broadcast_address(cidr):
         return str(broadcast)
 
 
+def cidr_broadcast_address_alternative(cidr):
+    """Wraps cidr_broadcast address and handles last cidr differently in
+    respect to netaddr
+
+    :param cidr: (string, netaddr.IPNetwork, netaddr.IPAddress) either an ipv4
+                 or ipv6 cidr.
+    :returns: (string) broadcast address of the cidr,
+              IP of the cidr if IPv4 /32 CIDR, otherwise None
+    """
+    broadcast = cidr_broadcast_address(cidr)
+    if broadcast:
+        return str(broadcast)
+    net = netaddr.IPNetwork(cidr)
+    if net.prefixlen == 32 and net.version == n_const.IP_VERSION_4:
+        return str(net.ip)
+    # NOTE(mtomaska): netaddr ver >= 0.9.0 will start returning None
+    # for IPv6 cidr /127 /128
+
+
 def get_ip_version(ip_or_cidr):
     return netaddr.IPNetwork(ip_or_cidr).version
 
 
 def ip_version_from_int(ip_version_int):
-    if ip_version_int == 4:
+    if ip_version_int == n_const.IP_VERSION_4:
         return n_const.IPv4
-    if ip_version_int == 6:
+    if ip_version_int == n_const.IP_VERSION_6:
         return n_const.IPv6
     raise ValueError(_('Illegal IP version number'))
 
@@ -337,8 +356,8 @@ def get_socket_address_family(ip_version):
 
 def is_version_greater_equal(version1, version2):
     """Returns True if version1 is greater or equal than version2 else False"""
-    return (pkg_resources.parse_version(version1) >=
-            pkg_resources.parse_version(version2))
+    return (versionutils.convert_version_to_tuple(version1) >=
+            versionutils.convert_version_to_tuple(version2))
 
 
 class DelayedStringRenderer(object):
@@ -645,7 +664,7 @@ def create_object_with_dependency(creator, dep_getter, dep_creator,
                     # sleep for a random time between 0 and 1 second to
                     # make sure a concurrent worker doesn't retry again
                     # at exactly the same time
-                    time.sleep(random.uniform(0, 1))
+                    time.sleep(secrets.SystemRandom().uniform(0, 1))
                     ctx.reraise = False
                     continue
         try:
@@ -745,7 +764,7 @@ def attach_exc_details(e, msg, args=_NO_ARGS_MARKER):
 def extract_exc_details(e):
     for attr in ('_error_context_msg', '_error_context_args'):
         if not hasattr(e, attr):
-            return u'No details.'
+            return 'No details.'
     details = e._error_context_msg
     args = e._error_context_args
     if args is _NO_ARGS_MARKER:
@@ -861,7 +880,8 @@ def port_ip_changed(new_port, original_port):
     return False
 
 
-def validate_rp_bandwidth(rp_bandwidths, device_names):
+def validate_rp_bandwidth(rp_bandwidths, device_names,
+                          tunnelled_network_rp_name=None):
     """Validate resource provider bandwidths against device names.
 
     :param rp_bandwidths: Dict containing resource provider bandwidths,
@@ -870,10 +890,14 @@ def validate_rp_bandwidth(rp_bandwidths, device_names):
     :param device_names: A set of the device names given in bridge_mappings
                          in case of ovs-agent or in physical_device_mappings
                          in case of sriov-agent
+    :param tunnelled_network_rp_name: the resource provider name for tunnelled
+                                      networks; if present, it will be added
+                                      to the devices list.
     :raises ValueError: In case of the devices (keys) in the rp_bandwidths dict
                         are not in the device_names set.
     """
-
+    if tunnelled_network_rp_name:
+        device_names.add(tunnelled_network_rp_name)
     for dev_name in rp_bandwidths:
         if dev_name not in device_names:
             raise ValueError(_(
@@ -929,7 +953,7 @@ def spawn_n(func, *args, **kwargs):
 def timecost(f):
     call_id = uuidutils.generate_uuid()
     message_base = ("Time-cost: call %(call_id)s function %(fname)s ") % {
-                    "call_id": call_id, "fname": f.__name__}
+        "call_id": call_id, "fname": f.__name__}
     end_message = (message_base + "took %(seconds).3fs seconds to run")
 
     @timeutils.time_it(LOG, message=end_message, min_duration=None)
@@ -1009,6 +1033,17 @@ def get_az_hints(resource):
             cfg.CONF.default_availability_zones)
 
 
+def disable_notifications(function):
+    """Decorator to disable notifications"""
+
+    @functools.wraps(function)
+    def wrapper(*args, **kwargs):
+        rpc_workers = conf_service.get_rpc_workers()
+        if rpc_workers is None or rpc_workers >= 1:
+            return function(*args, **kwargs)
+    return wrapper
+
+
 def skip_exceptions(exceptions):
     """Decorator to catch and hide any provided exception in the argument"""
 
@@ -1025,7 +1060,7 @@ def skip_exceptions(exceptions):
                 with excutils.save_and_reraise_exception() as ctx:
                     if issubclass(type(exc), tuple(exceptions)):
                         LOG.info('Skipped exception %s when calling method %s',
-                                 ctx.value.__repr__(), function.__repr__())
+                                 repr(ctx.value), repr(function))
                         ctx.reraise = False
         return wrapper
     return decorator
@@ -1041,3 +1076,30 @@ def effective_qos_policy_id(resource):
     """
     return (resource.get(qos_consts.QOS_POLICY_ID) or
             resource.get(qos_consts.QOS_NETWORK_POLICY_ID))
+
+
+def sign_instance_id(conf, instance_id):
+    secret = conf.metadata_proxy_shared_secret
+    secret = encodeutils.to_utf8(secret)
+    instance_id = encodeutils.to_utf8(instance_id)
+    return hmac.new(secret, instance_id, hashlib.sha256).hexdigest()
+
+
+def parse_permitted_ethertypes(permitted_ethertypes):
+    ret = set()
+    for pe in permitted_ethertypes:
+        if pe[:2].lower() != '0x':
+            LOG.warning('Custom ethertype %s is not a hexadecimal number.', pe)
+            continue
+
+        try:
+            value = hex(int(pe, 16))
+            if value in ret:
+                LOG.warning('Custom ethertype %s is repeated', pe)
+            else:
+                ret.add(value)
+        except ValueError:
+            LOG.warning('Custom ethertype %s is not a hexadecimal number.', pe)
+            continue
+
+    return ret
