@@ -66,7 +66,9 @@ class BaseFullStackTestCase(testlib_api.MySQLTestCaseMixin,
         # neutron server against this database.
         _orig_db_url = cfg.CONF.database.connection
         cfg.CONF.set_override(
-            'connection', str(self.engine.url), group='database')
+            'connection',
+            self.engine.url.render_as_string(hide_password=False),
+            group='database')
         self.addCleanup(
             cfg.CONF.set_override,
             "connection", _orig_db_url, group="database"
@@ -118,7 +120,7 @@ class BaseFullStackTestCase(testlib_api.MySQLTestCaseMixin,
         common_utils.wait_until_true(_agent_down)
 
     def _assert_ping_during_agents_restart(
-            self, agents, src_namespace, ips, restart_timeout=10,
+            self, agents, src_namespace, ips, restart_timeout=30,
             ping_timeout=1, count=10):
         with net_helpers.async_ping(
                 src_namespace, ips, timeout=ping_timeout,
@@ -130,7 +132,7 @@ class BaseFullStackTestCase(testlib_api.MySQLTestCaseMixin,
 
             futures.wait(restarts, timeout=restart_timeout)
 
-            self.assertTrue(all([r.done() for r in restarts]))
+            self.assertTrue(all(r.done() for r in restarts))
             LOG.debug("Restarting agents - done")
 
             # It is necessary to give agents time to initialize
@@ -165,11 +167,12 @@ class BaseFullStackTestCase(testlib_api.MySQLTestCaseMixin,
         available_ips = itertools.islice(valid_ips, initial, initial + num)
         return [str(available_ip) for available_ip in available_ips]
 
-    def _create_external_vm(self, network, subnet):
+    def _create_external_vm(self, network, subnet, ip=None):
+        ip = ip or subnet['gateway_ip']
         vm = self.useFixture(
             machine_fixtures.FakeMachine(
                 self.environment.central_bridge,
-                common_utils.ip_to_cidr(subnet['gateway_ip'], 24)))
+                common_utils.ip_to_cidr(ip, 24)))
         # NOTE(slaweq): as ext_net is 'vlan' network type external_vm needs to
         # send packets with proper vlan also
         vm.bridge.set_db_attribute(
