@@ -12,13 +12,17 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+import signal
+
 import fixtures
 from neutron_lib import constants
 from neutronclient.common import exceptions as nc_exc
 from oslo_config import cfg
+from oslo_log import log as logging
 
 from neutron.agent.linux import ip_lib
 from neutron.common import utils as common_utils
+from neutron.conf import quota as quota_conf
 from neutron.plugins.ml2.drivers.linuxbridge.agent import \
     linuxbridge_neutron_agent as lb_agent
 from neutron.tests.common.exclusive_resources import ip_address
@@ -27,6 +31,8 @@ from neutron.tests.common import net_helpers
 from neutron.tests.fullstack.resources import config
 from neutron.tests.fullstack.resources import process
 
+LOG = logging.getLogger(__name__)
+
 
 class EnvironmentDescription(object):
     """A set of characteristics of an environment setup.
@@ -41,7 +47,10 @@ class EnvironmentDescription(object):
                  debug_iptables=False, log=False, report_bandwidths=False,
                  has_placement=False, placement_port=None,
                  dhcp_scheduler_class=None, ml2_extension_drivers=None,
-                 api_workers=1):
+                 api_workers=1,
+                 enable_traditional_dhcp=True, local_ip_ext=False,
+                 quota_driver=quota_conf.QUOTA_DB_DRIVER,
+                 allow_experimental_linuxbridge=True):
         self.network_type = network_type
         self.l2_pop = l2_pop
         self.qos = qos
@@ -64,6 +73,12 @@ class EnvironmentDescription(object):
             self.service_plugins += ',log'
         self.ml2_extension_drivers = ml2_extension_drivers
         self.api_workers = api_workers
+        self.enable_traditional_dhcp = enable_traditional_dhcp
+        self.local_ip_ext = local_ip_ext
+        if self.local_ip_ext:
+            self.service_plugins += ',local_ip'
+        self.quota_driver = quota_driver
+        self.allow_experimental_linuxbridge = allow_experimental_linuxbridge
 
     @property
     def tunneling_enabled(self):
@@ -100,9 +115,8 @@ class Host(fixtures.Fixture):
     IP address on the appropriate physical NIC. The Host class does the same
     with the connect_* methods.
 
-    TODO(amuller): Add start/stop/restart methods that will start/stop/restart
-    all of the agents on this host. Add a kill method that stops all agents
-    and disconnects the host from other hosts.
+    TODO(amuller): Add restart method that will restart all of the agents on
+    this host.
     """
 
     def __init__(self, env_desc, host_desc, test_name,
@@ -253,12 +267,14 @@ class Host(fixtures.Fixture):
 
         veth_1.link.set_up()
         veth_2.link.set_up()
+        self.tunnel_device = veth_1
 
     def connect_to_central_network_via_vlans(self, host_data_bridge):
         # If using VLANs as a segmentation device, it's needed to connect
         # a provider bridge to a centralized, shared bridge.
-        net_helpers.create_patch_ports(
+        source, destination = net_helpers.create_patch_ports(
             self.central_bridge, host_data_bridge)
+        self.internal_port = destination
 
     def allocate_local_ip(self):
         if not self.env_desc.network_range:
@@ -286,6 +302,32 @@ class Host(fixtures.Fixture):
                 self.network_bridges[network_id] = bridge
         return bridge
 
+    def disconnect(self):
+        if self.env_desc.tunneling_enabled:
+            self.tunnel_device.addr.flush(4)
+        else:
+            self.br_phys.delete_port(self.internal_port)
+        LOG.info(f'Host {self.hostname} disconnected.')
+
+    def kill(self, parent=None):
+        # First kill all the agent to prevent a graceful shutdown
+        for agent_name, agent in self.agents.items():
+            agent.stop(kill_signal=signal.SIGKILL)
+        LOG.info(f'Agents on host {self.hostname} killed.')
+
+        self.shutdown(parent)
+
+    def shutdown(self, parent=None):
+        self.cleanUp()
+
+        # Remove cleanup function from parent because it can't be called twice
+        if parent:
+            parent._cleanups._cleanups.remove(
+                (self.cleanUp, (), {})
+            )
+
+        LOG.info(f'Host {self.hostname} shut down.')
+
     @property
     def hostname(self):
         return self.neutron_config.config.DEFAULT.host
@@ -375,6 +417,9 @@ class Environment(fixtures.Fixture):
         except nc_exc.NeutronClientException:
             return False
 
+    def get_host_by_name(self, hostname):
+        return next(host for host in self.hosts if host.hostname == hostname)
+
     def _create_host(self, host_desc):
         temp_dir = self.useFixture(fixtures.TempDir()).path
         neutron_config = config.NeutronConfigFixture(
