@@ -16,8 +16,11 @@ import functools
 import os
 import time
 
-import netaddr
+from datetime import datetime
+
 from neutron_lib import constants
+from neutronclient.common import exceptions
+from oslo_log import log as logging
 from oslo_utils import uuidutils
 
 from neutron.agent.l3 import ha_router
@@ -27,6 +30,7 @@ from neutron.agent.linux import l3_tc_lib
 from neutron.common import utils as common_utils
 from neutron.tests import base as tests_base
 from neutron.tests.common.exclusive_resources import ip_network
+from neutron.tests.common import net_helpers
 from neutron.tests.fullstack import base
 from neutron.tests.fullstack.resources import environment
 from neutron.tests.fullstack.resources import machine
@@ -34,6 +38,8 @@ from neutron.tests.unit import testlib_api
 
 load_tests = testlib_api.module_load_tests
 
+LOG = logging.getLogger(__name__)
+
 
 class TestL3Agent(base.BaseFullStackTestCase):
 
@@ -69,7 +75,8 @@ class TestL3Agent(base.BaseFullStackTestCase):
     def _boot_fake_vm_in_network(self, host, tenant_id, network_id, wait=True):
         vm = self.useFixture(
             machine.FakeFullstackMachine(
-                host, network_id, tenant_id, self.safe_client, use_dhcp=True))
+                host, network_id, tenant_id, self.safe_client,
+                use_dhcp=self.use_dhcp))
         if wait:
             vm.block_until_boot()
         return vm
@@ -174,10 +181,13 @@ class TestL3Agent(base.BaseFullStackTestCase):
         return "%s@%s" % (namespace, suffix)
 
     def _get_l3_agents_with_ha_state(
-            self, l3_agents, router_id, ha_state=None):
+            self, router_id, ha_state=None):
+        l3_agents = [host.agents['l3'] for host in self.environment.hosts
+                     if 'l3' in host.agents]
         found_agents = []
         agents_hosting_router = self.client.list_l3_agent_hosting_routers(
             router_id)['agents']
+
         for agent in l3_agents:
             agent_host = agent.neutron_cfg_fixture.get_host()
             for agent_hosting_router in agents_hosting_router:
@@ -188,6 +198,13 @@ class TestL3Agent(base.BaseFullStackTestCase):
                     break
         return found_agents
 
+    def _get_hosts_with_ha_state(
+            self, router_id, ha_state=None):
+        return [
+            self.environment.get_host_by_name(agent.hostname)
+            for agent in self._get_l3_agents_with_ha_state(router_id, ha_state)
+        ]
+
     def _router_fip_qos_after_admin_state_down_up(self, ha=False):
         def get_router_gw_interface():
             devices = ip.get_devices()
@@ -211,11 +228,9 @@ class TestL3Agent(base.BaseFullStackTestCase):
             tenant_id, 'fs_policy', 'Fullstack testing policy',
             shared='False', is_default='False')
         self.safe_client.create_bandwidth_limit_rule(
-            tenant_id, qos_policy['id'], 1111, 2222,
-            constants.INGRESS_DIRECTION)
+            qos_policy['id'], 1111, 2222, constants.INGRESS_DIRECTION)
         self.safe_client.create_bandwidth_limit_rule(
-            tenant_id, qos_policy['id'], 3333, 4444,
-            constants.EGRESS_DIRECTION)
+            qos_policy['id'], 3333, 4444, constants.EGRESS_DIRECTION)
 
         fip = self.safe_client.create_floatingip(
             tenant_id, ext_net['id'], vm.ip, vm.neutron_port['id'],
@@ -230,9 +245,7 @@ class TestL3Agent(base.BaseFullStackTestCase):
         external_vm.block_until_ping(fip['floating_ip_address'])
 
         if ha:
-            l3_agents = [host.agents['l3'] for host in self.environment.hosts]
-            router_agent = self._get_l3_agents_with_ha_state(
-                l3_agents, router['id'])[0]
+            router_agent = self._get_l3_agents_with_ha_state(router['id'])[0]
             qrouter_ns = self._get_namespace(
                             router['id'],
                             router_agent)
@@ -265,12 +278,53 @@ class TestL3Agent(base.BaseFullStackTestCase):
         return (_is_filter_set(constants.INGRESS_DIRECTION) and
                 _is_filter_set(constants.EGRESS_DIRECTION))
 
+    def _test_concurrent_router_subnet_attachment_overlapping_cidr(self,
+                                                                   ha=False):
+        tenant_id = uuidutils.generate_uuid()
+        subnet_cidr = '10.200.0.0/24'
+        # to have many port interactions where race conditions would happen
+        # deleting ports meanwhile find operations to evaluate the overlapping
+        subnets = 10
+
+        funcs = []
+        args = []
+        router = self.safe_client.create_router(tenant_id, ha=ha)
+
+        for i in range(subnets):
+            network_tmp = self.safe_client.create_network(
+                tenant_id, name='foo-network' + str(i))
+            subnet_tmp = self.safe_client.create_subnet(
+                tenant_id, network_tmp['id'], subnet_cidr)
+            funcs.append(self.safe_client.add_router_interface)
+            args.append((router['id'], subnet_tmp['id']))
+
+        exception_requests = self._simulate_concurrent_requests_process(
+            funcs, args)
+
+        if not all(isinstance(e, exceptions.BadRequest)
+                   for e in exception_requests):
+            self.fail('Unexpected exception adding interfaces to router from '
+                      'different subnets overlapping')
+
+        if len(exception_requests) < subnets - 1:
+            self.fail('If we have tried to associate %s subnets overlapping '
+                      'cidr to the router, we should have received at least '
+                      '%s or %s rejected requests, but we have only received '
+                      '%s' % (str(subnets), str(subnets - 1), str(subnets),
+                              str(len(exception_requests))))
+
 
 class TestLegacyL3Agent(TestL3Agent):
 
+    # NOTE(slaweq): don't use dhcp agents due to the oslo.privsep bug
+    # https://review.opendev.org/c/openstack/neutron/+/794994
+    # When it will be fixed DHCP can be used here again.
+    use_dhcp = False
+
     def setUp(self):
         host_descriptions = [
-            environment.HostDescription(l3_agent=True, dhcp_agent=True,
+            environment.HostDescription(l3_agent=True,
+                                        dhcp_agent=self.use_dhcp,
                                         l3_agent_extensions="fip_qos"),
             environment.HostDescription()]
         env = environment.Environment(
@@ -280,18 +334,6 @@ class TestLegacyL3Agent(TestL3Agent):
             host_descriptions)
         super(TestLegacyL3Agent, self).setUp(env)
 
-    def test_namespace_exists(self):
-        tenant_id = uuidutils.generate_uuid()
-
-        router = self.safe_client.create_router(tenant_id)
-        network = self.safe_client.create_network(tenant_id)
-        subnet = self.safe_client.create_subnet(
-            tenant_id, network['id'], '20.0.0.0/24', gateway_ip='20.0.0.1')
-        self.safe_client.add_router_interface(router['id'], subnet['id'])
-
-        namespace = self._get_namespace(router['id'])
-        self.assert_namespace_exists(namespace)
-
     def test_mtu_update(self):
         tenant_id = uuidutils.generate_uuid()
 
@@ -318,89 +360,6 @@ class TestLegacyL3Agent(TestL3Agent):
         network = self.safe_client.update_network(network['id'], mtu=mtu)
         common_utils.wait_until_true(lambda: ri_dev.link.mtu == mtu)
 
-    def test_east_west_traffic(self):
-        tenant_id = uuidutils.generate_uuid()
-        router = self.safe_client.create_router(tenant_id)
-
-        vm1 = self._create_net_subnet_and_vm(
-            tenant_id, ['20.0.0.0/24', '2001:db8:aaaa::/64'],
-            self.environment.hosts[0], router)
-        vm2 = self._create_net_subnet_and_vm(
-            tenant_id, ['21.0.0.0/24', '2001:db8:bbbb::/64'],
-            self.environment.hosts[1], router)
-
-        vm1.block_until_ping(vm2.ip)
-        # Verify ping6 from vm2 to vm1 IPv6 Address
-        vm2.block_until_ping(vm1.ipv6)
-
-    def test_north_south_traffic(self):
-        # This function creates an external network which is connected to
-        # central_bridge and spawns an external_vm on it.
-        # The external_vm is configured with the gateway_ip (both v4 & v6
-        # addresses) of external subnet. Later, it creates a tenant router,
-        # a tenant network and two tenant subnets (v4 and v6). The tenant
-        # router is associated with tenant network and external network to
-        # provide north-south connectivity to the VMs.
-        # We validate the following in this testcase.
-        # 1. SNAT support: using ping from tenant VM to external_vm
-        # 2. Floating IP support: using ping from external_vm to VM floating ip
-        # 3. IPv6 ext connectivity: using ping6 from tenant vm to external_vm.
-        tenant_id = uuidutils.generate_uuid()
-        ext_net, ext_sub = self._create_external_network_and_subnet(tenant_id)
-        external_vm = self._create_external_vm(ext_net, ext_sub)
-        # Create an IPv6 subnet in the external network
-        v6network = self.useFixture(
-            ip_network.ExclusiveIPNetwork(
-                "2001:db8:1234::1", "2001:db8:1234::10", "64")).network
-        ext_v6sub = self.safe_client.create_subnet(
-            tenant_id, ext_net['id'], v6network)
-
-        router = self.safe_client.create_router(tenant_id,
-                                                external_network=ext_net['id'])
-
-        # Configure the gateway_ip of external v6subnet on the external_vm.
-        external_vm.ipv6_cidr = common_utils.ip_to_cidr(
-            ext_v6sub['gateway_ip'], 64)
-
-        # Configure an IPv6 downstream route to the v6Address of router gw port
-        for fixed_ip in router['external_gateway_info']['external_fixed_ips']:
-            if netaddr.IPNetwork(fixed_ip['ip_address']).version == 6:
-                external_vm.set_default_gateway(fixed_ip['ip_address'])
-
-        vm = self._create_net_subnet_and_vm(
-            tenant_id, ['20.0.0.0/24', '2001:db8:aaaa::/64'],
-            self.environment.hosts[1], router)
-
-        # ping external vm to test snat
-        vm.block_until_ping(external_vm.ip)
-
-        fip = self.safe_client.create_floatingip(
-            tenant_id, ext_net['id'], vm.ip, vm.neutron_port['id'])
-
-        # ping floating ip from external vm
-        external_vm.block_until_ping(fip['floating_ip_address'])
-
-        # Verify VM is able to reach the router interface.
-        vm.block_until_ping(vm.gateway_ipv6)
-        # Verify north-south connectivity using ping6 to external_vm.
-        vm.block_until_ping(external_vm.ipv6)
-
-        # Now let's remove and create again phys bridge and check connectivity
-        # once again
-        br_phys = self.environment.hosts[0].br_phys
-        br_phys.destroy()
-        br_phys.create()
-        self.environment.hosts[0].connect_to_central_network_via_vlans(
-            br_phys)
-
-        # ping floating ip from external vm
-        external_vm.block_until_ping(fip['floating_ip_address'])
-
-        # Verify VM is able to reach the router interface.
-        vm.block_until_ping(vm.gateway_ipv6)
-        # Verify north-south connectivity using ping6 to external_vm.
-        vm.block_until_ping(external_vm.ipv6)
-
     def test_gateway_ip_changed(self):
         self._test_gateway_ip_changed()
 
@@ -410,17 +369,35 @@ class TestLegacyL3Agent(TestL3Agent):
     def test_router_fip_qos_after_admin_state_down_up(self):
         self._router_fip_qos_after_admin_state_down_up()
 
+    def test_concurrent_router_subnet_attachment_overlapping_cidr(self):
+        self._test_concurrent_router_subnet_attachment_overlapping_cidr()
+
 
 class TestHAL3Agent(TestL3Agent):
 
+    # NOTE(slaweq): don't use dhcp agents due to the oslo.privsep bug
+    # https://review.opendev.org/c/openstack/neutron/+/794994
+    # When it will be fixed DHCP can be used here again.
+    use_dhcp = False
+
     def setUp(self):
+        # Two hosts with L3 agent to host HA routers
         host_descriptions = [
-            environment.HostDescription(l3_agent=True, dhcp_agent=True,
+            environment.HostDescription(l3_agent=True,
+                                        dhcp_agent=self.use_dhcp,
                                         l3_agent_extensions="fip_qos")
             for _ in range(2)]
+
+        # Add two hosts for FakeFullstackMachines
+        host_descriptions.extend([
+            environment.HostDescription()
+            for _ in range(2)
+        ])
+
         env = environment.Environment(
             environment.EnvironmentDescription(
                 network_type='vlan', l2_pop=True,
+                agent_down_time=30,
                 qos=True),
             host_descriptions)
         super(TestHAL3Agent, self).setUp(env)
@@ -431,9 +408,6 @@ class TestHAL3Agent(TestL3Agent):
             agents['agents'][0]['ha_state'] != agents['agents'][1]['ha_state'])
 
     def test_ha_router(self):
-        # TODO(amuller): Test external connectivity before and after a
-        # failover, see: https://review.opendev.org/#/c/196393/
-
         tenant_id = uuidutils.generate_uuid()
         router = self.safe_client.create_router(tenant_id, ha=True)
 
@@ -449,6 +423,139 @@ class TestHAL3Agent(TestL3Agent):
                 router['id']),
             timeout=90)
 
+    def _test_ha_router_failover(self, method):
+        tenant_id = uuidutils.generate_uuid()
+
+        # Create router
+        router = self.safe_client.create_router(tenant_id, ha=True)
+        router_id = router['id']
+        agents = self.client.list_l3_agent_hosting_routers(router_id)
+        self.assertEqual(2, len(agents['agents']),
+                         'HA router must be scheduled to both nodes')
+
+        # Create internal subnet
+        network = self.safe_client.create_network(tenant_id)
+        subnet = self.safe_client.create_subnet(
+            tenant_id, network['id'], '20.0.0.0/24')
+        self.safe_client.add_router_interface(router_id, subnet['id'])
+
+        # Create external network
+        external_network = self.safe_client.create_network(
+            tenant_id, external=True)
+        self.safe_client.create_subnet(
+            tenant_id, external_network['id'], '42.0.0.0/24',
+            enable_dhcp=False)
+        self.safe_client.add_gateway_router(
+            router_id,
+            external_network['id'])
+
+        # Create internal VM
+        vm = self.useFixture(
+            machine.FakeFullstackMachine(
+                self.environment.hosts[2],
+                network['id'],
+                tenant_id,
+                self.safe_client))
+        vm.block_until_boot()
+
+        # Create external VM
+        external = self.useFixture(
+            machine.FakeFullstackMachine(
+                self.environment.hosts[3],
+                external_network['id'],
+                tenant_id,
+                self.safe_client))
+        external.block_until_boot()
+
+        common_utils.wait_until_true(
+            functools.partial(
+                self._is_ha_router_active_on_one_agent,
+                router_id),
+            timeout=90)
+
+        # Test external connectivity, failover, test again
+        pinger = net_helpers.Pinger(vm.namespace, external.ip, interval=0.1)
+        pinger.start()
+
+        # Ensure connectivity before disconnect
+        vm.block_until_ping(external.ip)
+
+        get_active_hosts = functools.partial(
+            self._get_hosts_with_ha_state,
+            router_id,
+            'active',
+        )
+
+        active_hosts = get_active_hosts()
+
+        # Only one host should be active
+        self.assertEqual(len(active_hosts), 1,
+                         'More than one active HA routers')
+
+        active_host = active_hosts[0]
+        backup_host = next(
+            h for h in self.environment.hosts if h != active_host)
+
+        start = datetime.now()
+
+        if method == 'disconnect':
+            active_host.disconnect()
+        elif method == 'kill':
+            active_host.kill(parent=self.environment)
+        elif method == 'shutdown':
+            active_host.shutdown(parent=self.environment)
+
+        if method != 'shutdown':
+            # Ensure connectivity is shortly lost if the failover is not
+            # graceful
+            vm.assert_no_ping(external.ip)
+
+        LOG.debug(f'Connectivity lost after {datetime.now() - start}')
+
+        # Ensure connectivity is restored
+        vm.block_until_ping(external.ip)
+        LOG.debug(f'Connectivity restored after {datetime.now() - start}')
+
+        # Assert the backup host got active
+        timeout = self.environment.env_desc.agent_down_time * 1.2
+        common_utils.wait_until_true(
+            lambda: backup_host in get_active_hosts(),
+            timeout=timeout,
+        )
+        LOG.debug(f'Active host asserted after {datetime.now() - start}')
+
+        if method in ('kill', 'shutdown'):
+            # Assert the previously active host is no longer active if it was
+            # killed or shutdown. In the disconnect case both hosts will stay
+            # active, but one host is disconnected from the data plane.
+            common_utils.wait_until_true(
+                lambda: active_host not in get_active_hosts(),
+                timeout=timeout,
+            )
+            LOG.debug(f'Inactive host asserted after {datetime.now() - start}')
+
+        # Stop probing processes
+        pinger.stop()
+
+        # With the default advert_int of 2s the keepalived master timeout is
+        # about 6s. Assert less than 90 lost packets (9 seconds)
+        threshold = 90
+
+        lost = pinger.sent - pinger.received
+        message = (f'Sent {pinger.sent} packets, received {pinger.received} '
+                   f'packets, lost {lost} packets')
+
+        self.assertLess(lost, threshold, message)
+
+    def test_ha_router_failover_graceful(self):
+        self._test_ha_router_failover('shutdown')
+
+    def test_ha_router_failover_host_failure(self):
+        self._test_ha_router_failover('kill')
+
+    def test_ha_router_failover_disconnect(self):
+        self._test_ha_router_failover('disconnect')
+
     def _get_keepalived_state(self, keepalived_state_file):
         with open(keepalived_state_file, "r") as fd:
             return fd.read()
@@ -535,11 +642,10 @@ class TestHAL3Agent(TestL3Agent):
         router_ip = router['external_gateway_info'][
             'external_fixed_ips'][0]['ip_address']
 
-        l3_agents = [host.agents['l3'] for host in self.environment.hosts]
         l3_standby_agents = self._get_l3_agents_with_ha_state(
-            l3_agents, router['id'], 'standby')
+            router['id'], 'standby')
         l3_active_agents = self._get_l3_agents_with_ha_state(
-            l3_agents, router['id'], 'active')
+            router['id'], 'active')
         self.assertEqual(1, len(l3_active_agents))
 
         # Let's check first if connectivity from external_vm to router's
@@ -560,3 +666,7 @@ class TestHAL3Agent(TestL3Agent):
 
     def test_router_fip_qos_after_admin_state_down_up(self):
         self._router_fip_qos_after_admin_state_down_up(ha=True)
+
+    def test_concurrent_router_subnet_attachment_overlapping_cidr(self):
+        self._test_concurrent_router_subnet_attachment_overlapping_cidr(
+            ha=True)
