@@ -26,6 +26,7 @@ from ovsdbapp.backend.ovs_idl import idlutils
 from neutron.common.ovn import constants as ovn_const
 from neutron.common.ovn import utils as ovn_utils
 from neutron.common import utils as n_utils
+from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf
 from neutron.scheduler import l3_ovn_scheduler as l3_sched
 from neutron.tests.functional import base
 from neutron.tests.functional.resources.ovsdb import events
@@ -40,10 +41,12 @@ class TestRouter(base.TestOVNFunctionalBase):
         self.chassis2 = self.add_fake_chassis(
             'ovs-host2', physical_nets=['physnet2', 'physnet3'],
             enable_chassis_as_gw=True, azs=[])
+        self.physnet_used = ['physnet1', 'physnet2', 'physnet3']
         self.cr_lrp_pb_event = events.WaitForCrLrpPortBindingEvent()
         self.sb_api.idl.notify_handler.watch_event(self.cr_lrp_pb_event)
 
-    def _create_router(self, name, gw_info=None, az_hints=None):
+    def _create_router(self, name, gw_info=None, az_hints=None,
+                       enable_ecmp=None, enable_bfd=None):
         router = {'router':
                   {'name': name,
                    'admin_state_up': True,
@@ -52,8 +55,22 @@ class TestRouter(base.TestOVNFunctionalBase):
             router['router']['availability_zone_hints'] = az_hints
         if gw_info:
             router['router']['external_gateway_info'] = gw_info
+        if enable_bfd:
+            router['router']['enable_default_route_bfd'] = enable_bfd
+        if enable_ecmp:
+            router['router']['enable_default_route_ecmp'] = enable_ecmp
         return self.l3_plugin.create_router(self.context, router)
 
+    def _add_external_gateways(self, router_id, external_gateways):
+        router = {'router': {'external_gateways': external_gateways}}
+        return self.l3_plugin.add_external_gateways(
+            self.context, router_id, body=router)
+
+    def _remove_external_gateways(self, router_id, external_gateways):
+        router = {'router': {'external_gateways': external_gateways}}
+        return self.l3_plugin.remove_external_gateways(
+            self.context, router_id, body=router)
+
     def _create_ext_network(self, name, net_type, physnet, seg,
                             gateway, cidr):
         arg_list = (pnet.NETWORK_TYPE, external_net.EXTERNAL,)
@@ -72,13 +89,55 @@ class TestRouter(base.TestOVNFunctionalBase):
                               ip_version=n_consts.IP_VERSION_4)
         return network
 
-    def _set_redirect_chassis_to_invalid_chassis(self, ovn_client):
+    def _unset_lrp_gw_chassis(self, ovn_client):
         with ovn_client._nb_idl.transaction(check_error=True) as txn:
-            for lrp in self.nb_api.tables[
-                    'Logical_Router_Port'].rows.values():
+            for lrp in self.nb_api.tables['Logical_Router_Port'].rows.values():
                 txn.add(ovn_client._nb_idl.update_lrouter_port(
-                    lrp.name,
-                    gateway_chassis=[ovn_const.OVN_GATEWAY_INVALID_CHASSIS]))
+                    lrp.name, gateway_chassis=[]))
+
+    def _get_gwc_dict(self):
+        sched_info = {}
+        for row in self.nb_api.db_list_rows("Logical_Router_Port").execute(
+                check_error=True):
+            for gwc in row.gateway_chassis:
+                chassis = sched_info.setdefault(gwc.chassis_name, {})
+                chassis[gwc.priority] = chassis.get(gwc.priority, 0) + 1
+        return sched_info
+
+    def _create_routers_wait_pb(self, begin, n, gw_info=None,
+                                bind_chassis=None):
+        routers = []
+        for i in range(begin, n + begin):
+            try:
+                router = self._create_router('router%d' % i, gw_info=gw_info)
+                routers.append(router)
+            except db_exc.DBReferenceError:
+                # NOTE(ralonsoh): this is a workaround for LP#1956344. There
+                # seems to be a bug in SQLite3. The "port" DB object is not
+                # persistently stored in the DB and raises a "DBReferenceError"
+                # exception occasionally.
+                continue
+
+            if bind_chassis:
+                gw_port_id = router.get('gw_port_id')
+                logical_port = 'cr-lrp-%s' % gw_port_id
+                self.assertTrue(self.cr_lrp_pb_event.wait(logical_port),
+                                msg='lrp %s failed to bind' % logical_port)
+                self.sb_api.lsp_bind(logical_port, bind_chassis,
+                                     may_exist=True).execute(check_error=True)
+        return routers
+
+    def _add_chassis(self, begin, n, physical_nets):
+        chassis_added = []
+        for i in range(begin, begin + n):
+            chassis_added.append(
+                self.add_fake_chassis(
+                    'ovs-host%s' % i,
+                    name=f'chassis-{i:02d}',
+                    physical_nets=physical_nets,
+                    other_config={
+                        'ovn-cms-options': 'enable-chassis-as-gw'}))
+        return chassis_added
 
     def test_gateway_chassis_on_router_gateway_port(self):
         ext2 = self._create_ext_network(
@@ -107,27 +166,26 @@ class TestRouter(base.TestOVNFunctionalBase):
         def fake_select(*args, **kwargs):
             self.assertCountEqual(candidates, kwargs['candidates'])
             # We are not interested in further processing, let us return
-            # INVALID_CHASSIS to avoid errors
-            return [ovn_const.OVN_GATEWAY_INVALID_CHASSIS]
+            # a random chassis name to avoid errors. If there are no
+            # candidates, this method returns None.
+            return ['a-random-chassis'] if candidates else None
 
-        with mock.patch.object(ovn_client._ovn_scheduler, 'select',
-                               side_effect=fake_select) as client_select,\
-            mock.patch.object(self.l3_plugin.scheduler, 'select',
-                              side_effect=fake_select) as plugin_select:
+        with mock.patch.object(self.l3_plugin.scheduler, 'select',
+                               side_effect=fake_select) as plugin_select:
             gw_info = {'network_id': ext1['network']['id']}
             self._create_router('router1', gw_info=gw_info,
                                 az_hints=router_az_hints)
-            self.assertFalse(plugin_select.called)
-            self.assertTrue(client_select.called)
-            client_select.reset_mock()
+            # If the network is tunnelled, the scheduler is not called.
+            check = self.assertTrue if physnet else self.assertFalse
+            check(plugin_select.called)
             plugin_select.reset_mock()
 
-            # set redirect-chassis to neutron-ovn-invalid-chassis, so
-            # that schedule_unhosted_gateways will try to schedule it
-            self._set_redirect_chassis_to_invalid_chassis(ovn_client)
+            # Unset the redirect-chassis so that schedule_unhosted_gateways
+            # will try to schedule it.
+            self._unset_lrp_gw_chassis(ovn_client)
             self.l3_plugin.schedule_unhosted_gateways()
-            self.assertFalse(client_select.called)
-            self.assertTrue(plugin_select.called)
+            check = self.assertTrue if candidates else self.assertFalse
+            check(plugin_select.called)
 
     def test_gateway_chassis_with_cms_and_bridge_mappings(self):
         # Both chassis1 and chassis3 are having proper bridge mappings,
@@ -148,11 +206,9 @@ class TestRouter(base.TestOVNFunctionalBase):
             'ext1', 'vlan', 'physnet1', 1, "10.0.0.1", "10.0.0.0/24")
         # As we have 'gateways' in the system, but without required
         # chassis we should not schedule gw in that case at all.
-        self._set_redirect_chassis_to_invalid_chassis(ovn_client)
-        with mock.patch.object(ovn_client._ovn_scheduler, 'select',
-                              return_value=[self.chassis1]), \
-            mock.patch.object(self.l3_plugin.scheduler, 'select',
-                              side_effect=[self.chassis1]):
+        self._unset_lrp_gw_chassis(ovn_client)
+        with mock.patch.object(self.l3_plugin.scheduler, 'select',
+                               side_effect=[self.chassis1]):
             gw_info = {'network_id': ext1['network']['id']}
             self._create_router('router1', gw_info=gw_info)
 
@@ -191,11 +247,9 @@ class TestRouter(base.TestOVNFunctionalBase):
             'ext1', 'vlan', 'physnet1', 1, "10.0.0.1", "10.0.0.0/24")
         # As we have 'gateways' in the system, but without required
         # chassis we should not schedule gw in that case at all.
-        self._set_redirect_chassis_to_invalid_chassis(ovn_client)
-        with mock.patch.object(ovn_client._ovn_scheduler, 'select',
-                              return_value=[self.chassis1]), \
-            mock.patch.object(self.l3_plugin.scheduler, 'select',
-                              side_effect=[self.chassis1]):
+        self._unset_lrp_gw_chassis(ovn_client)
+        with mock.patch.object(self.l3_plugin.scheduler, 'select',
+                               side_effect=[self.chassis1]):
             gw_info = {'network_id': ext1['network']['id']}
             self._create_router('router1', gw_info=gw_info, az_hints=['ovn'])
         self.del_fake_chassis(self.chassis3)
@@ -215,7 +269,7 @@ class TestRouter(base.TestOVNFunctionalBase):
     def test_gateway_chassis_no_physnet_tunnelled_network(self):
         # The GW network is tunnelled, no physnet defined --> no possible
         # candidates.
-        self._check_gateway_chassis_candidates([], physnet=None)
+        self._check_gateway_chassis_candidates(None, physnet=None)
 
     def test_gateway_chassis_least_loaded_scheduler(self):
         # This test will create 4 routers each with its own gateway.
@@ -249,6 +303,67 @@ class TestRouter(base.TestOVNFunctionalBase):
                 chassis[gwc.priority] = chassis.get(gwc.priority, 0) + 1
         self.assertEqual(expected, sched_info)
 
+    def test_gateway_chassis_least_loaded_scheduler_anti_affinity(self):
+        ovn_client = self.l3_plugin._ovn_client
+        ovn_client._ovn_scheduler = l3_sched.OVNGatewayLeastLoadedScheduler()
+        ext1 = self._create_ext_network(
+            'ext1', 'flat', 'physnet5', None, "10.10.50.1", "10.10.50.0/24")
+        gw_info = {'network_id': ext1['network']['id']}
+
+        chassis_list = []
+        # first fill a few chassis with normal routers
+        chassis_list.extend(
+            self._add_chassis(0, ovn_const.MAX_GW_CHASSIS * 2, ['physnet5']))
+        for i in range(1, (ovn_const.MAX_GW_CHASSIS * 4) + 1):
+            router = self._create_router('router%d' % i, gw_info=gw_info)
+
+        # add more chassis and create a set of routers with multiple gateway
+        # ports
+        #
+        # This will stage a situation where a few chassis have higher load
+        # which we can use to confirm that the anti-affinity algorithm works as
+        # expected.
+        #
+        # Each router created below will have three LRPs, which should fit
+        # in ovn_const.MAX_GW_CHASSIS * 3 chassis without duplicates when
+        # using the anti affinity scheduler.
+        chassis_list.extend(
+            self._add_chassis(
+                len(chassis_list), ovn_const.MAX_GW_CHASSIS, ['physnet5']))
+        router_lrps = {}
+        for i in range(1, 2 + 1):
+            router = self._create_router('router-multi-gw%d' % i)
+            router_lrps[router['id']] = []
+            self._add_external_gateways(
+                router['id'],
+                [
+                    {'network_id': ext1['network']['id']},
+                    {'network_id': ext1['network']['id']},
+                    {'network_id': ext1['network']['id']},
+                ])
+            for row in self.nb_api.tables[
+                    'Logical_Router_Port'].rows.values():
+                if (ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY
+                        not in row.external_ids):
+                    continue
+                if row.external_ids[
+                        ovn_const.OVN_ROUTER_NAME_EXT_ID_KEY] == router['id']:
+                    chassis = {}
+                    for gwc in row.gateway_chassis:
+                        chassis[gwc.priority] = gwc.chassis_name
+                    router_lrps[router['id']].append(chassis)
+        for router, lrp_lists in router_lrps.items():
+            while lrp_lists:
+                try:
+                    lrps = lrp_lists.pop()
+                except IndexError:
+                    break
+                for lrp_list in lrp_lists:
+                    for n in range(1, ovn_const.MAX_GW_CHASSIS + 1):
+                        self.assertNotEqual(
+                            lrps[n],
+                            lrp_list[n])
+
     def _get_gw_port(self, router_id):
         router = self.l3_plugin._get_router(self.context, router_id)
         gw_port_id = router.get('gw_port_id', '')
@@ -262,13 +377,8 @@ class TestRouter(base.TestOVNFunctionalBase):
         Test cases when subnets are added to an external network after router
         has been configured to use that network via "set --external-gateway"
         """
-
-        ovn_client = self.l3_plugin._ovn_client
-
-        with mock.patch.object(
-                ovn_client._ovn_scheduler, 'select',
-                return_value=[ovn_const.OVN_GATEWAY_INVALID_CHASSIS]) as \
-                client_select:
+        with mock.patch.object(self.l3_plugin.scheduler, 'select',
+                               return_value=self.chassis1) as plugin_select:
             router1 = self._create_router('router1', gw_info=None)
             router_id = router1['id']
             self.assertIsNone(self._get_gw_port(router_id),
@@ -306,7 +416,7 @@ class TestRouter(base.TestOVNFunctionalBase):
             self.assertIsNotNone(self._get_gw_port(router_id),
                                  "router logical port must exist after v6 add")
 
-            self.assertGreaterEqual(client_select.call_count, 3)
+            self.assertGreaterEqual(plugin_select.call_count, 3)
 
             # Verify that ports have had the subnets created
             kwargs = {'device_owner': n_consts.DEVICE_OWNER_ROUTER_GW}
@@ -382,20 +492,18 @@ class TestRouter(base.TestOVNFunctionalBase):
         def fake_select(*args, **kwargs):
             self.assertCountEqual(self.candidates, kwargs['candidates'])
             # We are not interested in further processing, let us return
-            # INVALID_CHASSIS to avoid errors
-            return [ovn_const.OVN_GATEWAY_INVALID_CHASSIS]
+            # a random chassis name to avoid errors.
+            return ['a-random-chassis']
 
-        with mock.patch.object(ovn_client._ovn_scheduler, 'select',
-                               side_effect=fake_select) as client_select,\
-            mock.patch.object(self.l3_plugin.scheduler, 'select',
-                              side_effect=fake_select) as plugin_select:
+        with mock.patch.object(self.l3_plugin.scheduler, 'select',
+                               side_effect=fake_select) as plugin_select:
             self.candidates = [self.chassis1]
             gw_info = {'network_id': ext1['network']['id']}
             router1 = self._create_router('router1', gw_info=gw_info)
 
-            # set redirect-chassis to neutron-ovn-invalid-chassis, so
-            # that schedule_unhosted_gateways will try to schedule it
-            self._set_redirect_chassis_to_invalid_chassis(ovn_client)
+            # Unset the redirect-chassis so that schedule_unhosted_gateways
+            # will try to schedule it.
+            self._unset_lrp_gw_chassis(ovn_client)
             self.l3_plugin.schedule_unhosted_gateways()
 
             self.candidates = [self.chassis1, self.chassis2]
@@ -403,7 +511,7 @@ class TestRouter(base.TestOVNFunctionalBase):
             self.l3_plugin.update_router(
                 self.context, router1['id'],
                 {'router': {l3_apidef.EXTERNAL_GW_INFO: gw_info}})
-            self._set_redirect_chassis_to_invalid_chassis(ovn_client)
+            self._unset_lrp_gw_chassis(ovn_client)
             self.l3_plugin.schedule_unhosted_gateways()
 
             self.candidates = []
@@ -411,25 +519,17 @@ class TestRouter(base.TestOVNFunctionalBase):
             self.l3_plugin.update_router(
                 self.context, router1['id'],
                 {'router': {l3_apidef.EXTERNAL_GW_INFO: gw_info}})
-            self._set_redirect_chassis_to_invalid_chassis(ovn_client)
+            self._unset_lrp_gw_chassis(ovn_client)
             self.l3_plugin.schedule_unhosted_gateways()
 
             # We can't test call_count for these mocks, as we have disabled
             # maintenance_worker which will trigger chassis events
             # and eventually calling schedule_unhosted_gateways.
-            # However, we know for sure that these mocks must have been
-            # called at least 3 times because that is the number of times
-            # this test invokes them: 1x create_router + 2x update_router
-            # for client_select mock; and 3x schedule_unhosted_gateways for
-            # plugin_select mock.
-            self.assertGreaterEqual(client_select.call_count, 3)
-            self.assertGreaterEqual(plugin_select.call_count, 3)
-
-    def _find_port_binding(self, port_id):
-        cmd = self.sb_api.db_find_rows('Port_Binding',
-                                       ('logical_port', '=', port_id))
-        rows = cmd.execute(check_error=True)
-        return rows[0] if rows else None
+            # The router is created with a gateway port and updated twice.
+            # However, the "plugin_select" is called only twice because the
+            # third gateway network used is type "geneve" and the LRP are not
+            # hosted in any chassis.
+            self.assertGreaterEqual(plugin_select.call_count, 2)
 
     def test_router_gateway_port_binding_host_id(self):
         # Test setting chassis on chassisredirect port in Port_Binding table,
@@ -453,30 +553,14 @@ class TestRouter(base.TestOVNFunctionalBase):
                              may_exist=True).execute(check_error=True)
 
         def check_port_binding_host_id(port_id):
-            # Get port from Neutron DB
             port = core_plugin.get_ports(
                 self.context, filters={'id': [port_id]})[0]
-            # Get port from OVN DB
-            bp = self._find_port_binding(port_id)
-            ovn_host_id = bp.external_ids.get(ovn_const.OVN_HOST_ID_EXT_ID_KEY)
-            return port[portbindings.HOST_ID] == host_id == ovn_host_id
+            return port[portbindings.HOST_ID] == host_id
 
         # Test if router gateway port updated with this chassis
         n_utils.wait_until_true(lambda: check_port_binding_host_id(
             gw_port_id))
 
-        # Simulate failover to another chassis and check host_id in Neutron DB
-        # and external_ids:neutron:host_id in OVN DB are updated
-        chassis = idlutils.row_by_value(
-            self.sb_api.idl, "Chassis", "name", self.chassis2
-        )
-        host_id = chassis.hostname
-        self.sb_api.lsp_unbind(logical_port).execute(check_error=True)
-        self.sb_api.lsp_bind(logical_port, self.chassis2).execute(
-            check_error=True
-        )
-        n_utils.wait_until_true(lambda: check_port_binding_host_id(gw_port_id))
-
     def _validate_router_ipv6_ra_configs(self, lrp_name, expected_ra_confs):
         lrp = idlutils.row_by_value(self.nb_api.idl,
                                     'Logical_Router_Port', 'name', lrp_name)
@@ -529,16 +613,151 @@ class TestRouter(base.TestOVNFunctionalBase):
         self._test_router_port_ipv6_ra_configs_helper(
             ip_version=4)
 
-    def test_gateway_chassis_rebalance(self):
-        def _get_result_dict():
-            sched_info = {}
-            for row in self.nb_api.tables[
-                    'Logical_Router_Port'].rows.values():
-                for gwc in row.gateway_chassis:
-                    chassis = sched_info.setdefault(gwc.chassis_name, {})
-                    chassis[gwc.priority] = chassis.get(gwc.priority, 0) + 1
-            return sched_info
+    def test_create_delete_router_multiple_gw_ports(self):
+        ext4 = self._create_ext_network(
+            'ext4', 'flat', self.physnet_used[0], None, '40.0.0.1',
+            '40.0.0.0/24')
+        router = self._create_router('router4')
+        gws = self._add_external_gateways(
+            router['id'],
+            [
+                {'network_id': ext4['network']['id']},
+                {'network_id': ext4['network']['id']},
+            ]
+        )
+        lr = self.nb_api.lookup('Logical_Router',
+                                ovn_utils.ovn_name(router['id']))
+        self.assertEqual(
+            len(lr.ports),
+            len(gws['router']['external_gateways']))
 
+        self.assertEqual(
+            len(lr.static_routes),
+            1)
+
+        self.l3_plugin.delete_router(self.context, id=router['id'])
+        self.assertRaises(idlutils.RowNotFound, self.nb_api.lookup,
+                          'Logical_Router', ovn_utils.ovn_name(router['id']))
+
+    def test_create_router_multiple_gw_ports_ecmp(self):
+        ext5 = self._create_ext_network(
+            'ext5', 'flat', self.physnet_used[1], None, '10.0.50.1',
+            '10.0.50.0/24')
+        router = self._create_router('router5', enable_ecmp=True)
+        gws = self._add_external_gateways(
+            router['id'],
+            [
+                {'network_id': ext5['network']['id']},
+                {'network_id': ext5['network']['id']},
+            ]
+        )
+        lr = self.nb_api.lookup('Logical_Router',
+                                ovn_utils.ovn_name(router['id']))
+        # Check that the expected number of ports are created
+        self.assertEqual(
+            len(lr.ports),
+            len(gws['router']['external_gateways']))
+        # Check that the expected number of static routes are created
+        self.assertEqual(
+            len(lr.static_routes),
+            len(gws['router']['external_gateways']))
+
+    def test_create_delete_router_multiple_gw_ports_ecmp_and_bfd(self):
+        default_gw = "10.0.60.1"
+        ext6 = self._create_ext_network(
+            'ext6', 'flat', 'physnet6', None, default_gw, "10.0.60.0/24")
+        router = self._create_router('router6', gw_info=None,
+                                     enable_bfd=True, enable_ecmp=True)
+        gws = self._add_external_gateways(
+            router['id'],
+            [
+                {'network_id': ext6['network']['id']},
+                {'network_id': ext6['network']['id']},
+            ])
+        lr = self.nb_api.lookup('Logical_Router',
+                                ovn_utils.ovn_name(router['id']))
+
+        # Check that the expected number of ports are created
+        self.assertEqual(
+            len(lr.ports),
+            len(gws['router']['external_gateways']))
+        # Check that the expected number of static routes are created
+        self.assertEqual(
+            len(lr.static_routes),
+            len(gws['router']['external_gateways']))
+        # Check that static_route bfd and output_port attributes is set to the
+        # expected values
+        for static_route in lr.static_routes:
+            self.assertNotEqual(
+                [],
+                static_route.bfd)
+            self.assertNotEqual(
+                [],
+                static_route.output_port)
+            self.assertIn(static_route.output_port[0],
+                          [lrp.name
+                           for lrp in lr.ports])
+            self.assertIn(static_route.bfd[0].logical_port,
+                          [lrp.name
+                           for lrp in lr.ports])
+            self.assertEqual(static_route.bfd[0].logical_port,
+                             static_route.output_port[0])
+
+        router_ips = set()
+        for ext_gws in gws['router']['external_gateways']:
+            for ext_fip in ext_gws['external_fixed_ips']:
+                router_ips.add(ext_fip['ip_address'])
+
+        lrps = set()
+        for lrp in lr.ports:
+            for network in lrp.networks:
+                self.assertIn(
+                    network.split('/')[0],
+                    router_ips)
+            lrps.add(lrp.name)
+            bfd_rows = self.nb_api.bfd_find(
+                    lrp.name, default_gw).execute(check_error=True)
+            if not bfd_rows:
+                raise AssertionError('None of the expected BFD rows found.')
+            for bfd_row in bfd_rows:
+                self.assertEqual(
+                    bfd_row.logical_port,
+                    lrp.name)
+                self.assertEqual(
+                    bfd_row.dst_ip,
+                    default_gw)
+
+        self.l3_plugin.delete_router(self.context, id=router['id'])
+        self.assertRaises(idlutils.RowNotFound, self.nb_api.lookup,
+                          'Logical_Router', ovn_utils.ovn_name(router['id']))
+        for lrp_name in lrps:
+            if self.nb_api.bfd_find(
+                    lrp.name, default_gw).execute(check_error=True):
+                raise AssertionError('Unexpectedly found BFD rows.')
+
+    def test_update_router_single_gw_bfd(self):
+        ext1 = self._create_ext_network(
+            'ext7', 'flat', 'physnet1', None, "10.0.70.1", "10.0.70.0/24")
+        gw_info = {'network_id': ext1['network']['id']}
+        router = self._create_router('router7', gw_info=gw_info)
+        self.assertFalse(router['enable_default_route_bfd'])
+        lr = self.nb_api.lr_get(ovn_utils.ovn_name(router['id'])).execute()
+        for route in ovn_utils.get_lrouter_ext_gw_static_route(lr):
+            self.assertEqual(
+                [],
+                route.bfd)
+
+        router = self.l3_plugin.update_router(
+            self.context, router['id'],
+            {'router': {'enable_default_route_bfd': True}})
+        self.assertTrue(router['enable_default_route_bfd'])
+        lr = self.nb_api.lr_get(ovn_utils.ovn_name(router['id'])).execute()
+        for route in ovn_utils.get_lrouter_ext_gw_static_route(lr):
+            self.assertNotEqual(
+                [],
+                route.bfd)
+
+    def test_gateway_chassis_rebalance(self):
         ovn_client = self.l3_plugin._ovn_client
         chassis4 = self.add_fake_chassis(
             'ovs-host4', physical_nets=['physnet4'], other_config={
@@ -549,27 +768,11 @@ class TestRouter(base.TestOVNFunctionalBase):
         gw_info = {'network_id': ext1['network']['id']}
         # Tries to create 5 routers with a gateway. Since we're using
         # physnet4, the chassis candidates will be chassis4 initially.
-        num_routers = 5
-        for i in range(num_routers):
-            try:
-                router = self._create_router('router%d' % i, gw_info=gw_info)
-            except db_exc.DBReferenceError:
-                # NOTE(ralonsoh): this is a workaround for LP#1956344. There
-                # seems to be a bug in SQLite3. The "port" DB object is not
-                # persistently stored in the DB and raises a "DBReferenceError"
-                # exception occasionally.
-                num_routers -= 1
-                continue
-
-            gw_port_id = router.get('gw_port_id')
-            logical_port = 'cr-lrp-%s' % gw_port_id
-            self.assertTrue(self.cr_lrp_pb_event.wait(logical_port),
-                            msg='lrp %s failed to bind' % logical_port)
-            self.sb_api.lsp_bind(logical_port, chassis4,
-                                 may_exist=True).execute(check_error=True)
+        num_routers = len(self._create_routers_wait_pb(
+            1, 20, gw_info=gw_info, bind_chassis=chassis4))
         self.l3_plugin.schedule_unhosted_gateways()
         expected = {chassis4: {1: num_routers}}
-        self.assertEqual(expected, _get_result_dict())
+        self.assertEqual(expected, self._get_gwc_dict())
 
         # Add another chassis as a gateway chassis
         chassis5 = self.add_fake_chassis(
@@ -583,9 +786,9 @@ class TestRouter(base.TestOVNFunctionalBase):
 
         # Chassis4 should have all ports at Priority 2
         self.l3_plugin.schedule_unhosted_gateways()
-        self.assertEqual({2: num_routers}, _get_result_dict()[chassis4])
+        self.assertEqual({2: num_routers}, self._get_gwc_dict()[chassis4])
         # Chassis5 should have all ports at Priority 1
-        self.assertEqual({1: num_routers}, _get_result_dict()[chassis5])
+        self.assertEqual({1: num_routers}, self._get_gwc_dict()[chassis5])
 
         # delete chassis that hosts all the gateways
         self.del_fake_chassis(chassis4)
@@ -594,7 +797,7 @@ class TestRouter(base.TestOVNFunctionalBase):
         # As Chassis4 has been removed so all gateways that were
         # hosted there are now primaries on chassis5 and have
         # priority 1.
-        self.assertEqual({1: num_routers}, _get_result_dict()[chassis5])
+        self.assertEqual({1: num_routers}, self._get_gwc_dict()[chassis5])
 
     def test_gateway_chassis_rebalance_max_chassis(self):
         chassis_list = []
@@ -621,3 +824,52 @@ class TestRouter(base.TestOVNFunctionalBase):
                 'Logical_Router_Port'].rows.values():
             self.assertEqual(ovn_const.MAX_GW_CHASSIS,
                              len(row.gateway_chassis))
+
+    def test_set_router_mac_age_limit(self):
+        name = "macage_router1"
+        router = self._create_router(name)
+        lr_name = ovn_utils.ovn_name(router['id'])
+        options = self.nb_api.db_get(
+            "Logical_Router", lr_name, "options").execute(check_error=True)
+        self.assertEqual(ovn_conf.get_ovn_mac_binding_age_threshold(),
+                         options[ovn_const.LR_OPTIONS_MAC_AGE_LIMIT])
+
+    def test_schedule_unhosted_gateways_single_transaction(self):
+        ext1 = self._create_ext_network(
+            'ext1', 'flat', 'physnet6', None, "10.0.60.1", "10.0.60.0/24")
+        gw_info = {'network_id': ext1['network']['id']}
+
+        # Attempt to add 4 routers, since there are no chassis, none of them
+        # will be scheduled on any chassis.
+        num_routers = len(self._create_routers_wait_pb(1, 4, gw_info=gw_info))
+        self.assertEqual({}, self._get_gwc_dict())
+
+        # Add 2 chassis and rebalance gateways.
+        #
+        # The ovsdb_monitor.ChassisEvent handler will attempt to schedule
+        # unhosted gateways as chassis are added.
+        #
+        # Temporarily mock it out while adding the chassis so that we can get a
+        # predictable result for the purpose of this test.
+        chassis_list = []
+        with mock.patch.object(
+                self.l3_plugin, 'schedule_unhosted_gateways'):
+            chassis_list.extend(self._add_chassis(1, 2, ['physnet6']))
+        self.assertEqual({}, self._get_gwc_dict())
+
+        # Wrap `self.l3_plugin._nb_ovn.transaction` so that we can assert on
+        # number of calls.
+        with mock.patch.object(
+                self.l3_plugin._nb_ovn, 'transaction',
+                wraps=self.l3_plugin._nb_ovn.transaction) as wrap_txn:
+            self.l3_plugin.schedule_unhosted_gateways()
+            # The server is alive and we can't control the exact number of
+            # calls made to `_nb_ovn.transaction`. We can however check that
+            # the number of calls is less than number of unhosted gateways.
+            self.assertLess(len(wrap_txn.mock_calls), num_routers)
+
+        # Ensure the added gateways are spread evenly on the added chassis.
+        self.assertEqual(
+            {'chassis-01': {1: 2, 2: 2},
+             'chassis-02': {1: 2, 2: 2}},
+            self._get_gwc_dict())
