@@ -25,10 +25,13 @@ from oslo_utils import uuidutils
 
 from neutron.agent.l3 import agent as l3_agent
 from neutron.agent.l3 import router_info
+from neutron.agent.linux import external_process as ep
 from neutron.agent.linux import ip_lib
 from neutron.agent.linux import iptables_manager
 from neutron.agent.linux import utils as linux_utils
 from neutron.agent.metadata import driver as metadata_driver
+from neutron.agent.metadata import driver_base
+from neutron.common import metadata as comm_meta
 from neutron.conf.agent import common as agent_config
 from neutron.conf.agent.l3 import config as l3_config
 from neutron.conf.agent.l3 import ha as ha_conf
@@ -52,14 +55,14 @@ class TestMetadataDriverRules(base.BaseTestCase):
                  '-p tcp -m tcp --dport 80 -j REDIRECT --to-ports 9697')
         self.assertEqual(
             [rules],
-            metadata_driver.MetadataDriver.metadata_nat_rules(9697))
+            metadata_driver.metadata_nat_rules(9697))
 
     def test_metadata_nat_rules_ipv6(self):
         rules = ('PREROUTING', '-d fe80::a9fe:a9fe/128 -i qr-+ '
                  '-p tcp -m tcp --dport 80 -j REDIRECT --to-ports 9697')
         self.assertEqual(
             [rules],
-            metadata_driver.MetadataDriver.metadata_nat_rules(
+            metadata_driver.metadata_nat_rules(
                 9697, metadata_address='fe80::a9fe:a9fe/128'))
 
     def test_metadata_filter_rules(self):
@@ -68,7 +71,7 @@ class TestMetadataDriverRules(base.BaseTestCase):
                  ('INPUT', '-p tcp -m tcp --dport 9697 -j DROP')]
         self.assertEqual(
             rules,
-            metadata_driver.MetadataDriver.metadata_filter_rules(9697, '0x1'))
+            metadata_driver.metadata_filter_rules(9697, '0x1'))
 
 
 class TestMetadataDriverProcess(base.BaseTestCase):
@@ -80,6 +83,12 @@ class TestMetadataDriverProcess(base.BaseTestCase):
     METADATA_PORT = 8080
     METADATA_SOCKET = '/socket/path'
     PIDFILE = 'pidfile'
+    RATE_LIMIT_CONFIG = {
+        'base_window_duration': 10,
+        'base_query_rate_limit': 5,
+        'burst_window_duration': 1,
+        'burst_query_rate_limit': 10,
+    }
 
     def setUp(self):
         super(TestMetadataDriverProcess, self).setUp()
@@ -92,7 +101,7 @@ class TestMetadataDriverProcess(base.BaseTestCase):
         mock.patch('neutron.agent.l3.ha.AgentMixin'
                    '._init_ha_conf_path').start()
         self.delete_if_exists = mock.patch.object(linux_utils,
-                                                  'delete_if_exists')
+                                                  'delete_if_exists').start()
         self.mock_get_process = mock.patch.object(
             metadata_driver.MetadataDriver,
             '_get_metadata_proxy_process_manager')
@@ -100,6 +109,9 @@ class TestMetadataDriverProcess(base.BaseTestCase):
         l3_config.register_l3_agent_config_opts(l3_config.OPTS, cfg.CONF)
         ha_conf.register_l3_agent_ha_opts()
         meta_conf.register_meta_conf_opts(meta_conf.SHARED_OPTS, cfg.CONF)
+        meta_conf.register_meta_conf_opts(
+            meta_conf.METADATA_RATE_LIMITING_OPTS, cfg.CONF,
+            group=meta_conf.RATE_LIMITING_GROUP)
 
     def test_after_router_updated_called_on_agent_process_update(self):
         with mock.patch.object(metadata_driver, 'after_router_updated') as f,\
@@ -141,9 +153,10 @@ class TestMetadataDriverProcess(base.BaseTestCase):
             agent._process_updated_router(router)
             f.assert_not_called()
 
-    def _test_spawn_metadata_proxy(self, dad_failed=False):
+    def _test_spawn_metadata_proxy(self, dad_failed=False, rate_limited=False):
         router_id = _uuid()
         router_ns = 'qrouter-%s' % router_id
+        service_name = 'haproxy'
         ip_class_path = 'neutron.agent.linux.ip_lib.IPWrapper'
 
         cfg.CONF.set_override('metadata_proxy_user', self.EUNAME)
@@ -201,12 +214,13 @@ class TestMetadataDriverProcess(base.BaseTestCase):
                 bind_interface='fake-if')
 
             netns_execute_args = [
-                'haproxy',
+                service_name,
                 '-f', cfg_file]
 
-            log_tag = ("haproxy-" + metadata_driver.METADATA_SERVICE_NAME +
+            log_tag = ("haproxy-" + driver_base.METADATA_SERVICE_NAME +
                        "-" + router_id)
-            cfg_contents = metadata_driver._HAPROXY_CONFIG_TEMPLATE % {
+
+            expected_params = {
                 'user': self.EUNAME,
                 'group': self.EGNAME,
                 'host': self.METADATA_DEFAULT_IP,
@@ -227,27 +241,60 @@ class TestMetadataDriverProcess(base.BaseTestCase):
             else:
                 mock_del.assert_not_called()
 
+            if rate_limited:
+                expected_params.update(self.RATE_LIMIT_CONFIG,
+                                       stick_table_expire=10,
+                                       ip_version='ip')
+                expected_config_template = (
+                    comm_meta.METADATA_HAPROXY_GLOBAL +
+                    comm_meta.RATE_LIMITED_CONFIG_TEMPLATE +
+                    metadata_driver._HEADER_CONFIG_TEMPLATE)
+            else:
+                expected_config_template = (
+                    comm_meta.METADATA_HAPROXY_GLOBAL +
+                    driver_base._UNLIMITED_CONFIG_TEMPLATE +
+                    metadata_driver._HEADER_CONFIG_TEMPLATE)
+
             mock_open.assert_has_calls([
                 mock.call(cfg_file, 'w'),
-                mock.call().write(cfg_contents)], any_order=True)
+                mock.call().write(expected_config_template %
+                                  expected_params)], any_order=True)
 
+            env = {ep.PROCESS_TAG: service_name + '-' + router_id}
             ip_mock.assert_has_calls([
                 mock.call(namespace=router_ns),
-                mock.call().netns.execute(netns_execute_args, addl_env=None,
+                mock.call().netns.execute(netns_execute_args, addl_env=env,
                                           run_as_root=True)
             ])
 
             agent.process_monitor.register.assert_called_once_with(
-                router_id, metadata_driver.METADATA_SERVICE_NAME,
+                router_id, driver_base.METADATA_SERVICE_NAME,
                 mock.ANY)
 
+            self.delete_if_exists.assert_called_once_with(
+                mock.ANY, run_as_root=True)
+
     def test_spawn_metadata_proxy(self):
         self._test_spawn_metadata_proxy()
 
+    def test_spawn_rate_limited_metadata_proxy(self):
+        cfg.CONF.set_override('rate_limit_enabled', True,
+                              group=meta_conf.RATE_LIMITING_GROUP)
+        for k, v in self.RATE_LIMIT_CONFIG.items():
+            cfg.CONF.set_override(k, v, group=meta_conf.RATE_LIMITING_GROUP)
+
+        return self._test_spawn_metadata_proxy(rate_limited=True)
+
+    def test_metadata_proxy_conf_parse_ip_versions(self):
+        self.assertEqual(4, comm_meta.parse_ip_versions([4]))
+        self.assertEqual(6, comm_meta.parse_ip_versions([6]))
+        self.assertIsNone(comm_meta.parse_ip_versions([4, 6]))
+        self.assertIsNone(comm_meta.parse_ip_versions([5, 6]))
+
     def test_spawn_metadata_proxy_dad_failed(self):
         self._test_spawn_metadata_proxy(dad_failed=True)
 
-    @mock.patch.object(metadata_driver.LOG, 'error')
+    @mock.patch.object(driver_base.LOG, 'error')
     def test_spawn_metadata_proxy_handles_process_exception(self, error_log):
         process_instance = mock.Mock(active=False)
         process_instance.enable.side_effect = (
@@ -274,8 +321,9 @@ class TestMetadataDriverProcess(base.BaseTestCase):
                                                          mock.ANY, mock.ANY,
                                                          self.EUNAME,
                                                          self.EGNAME,
-                                                         mock.ANY, mock.ANY)
-            self.assertRaises(metadata_driver.InvalidUserOrGroupException,
+                                                         mock.ANY, mock.ANY,
+                                                         mock.ANY)
+            self.assertRaises(comm_meta.InvalidUserOrGroupException,
                               config.create_config_file)
 
     def test_create_config_file_wrong_group(self):
@@ -287,14 +335,13 @@ class TestMetadataDriverProcess(base.BaseTestCase):
                                                          mock.ANY, mock.ANY,
                                                          self.EUNAME,
                                                          self.EGNAME,
-                                                         mock.ANY, mock.ANY)
-            self.assertRaises(metadata_driver.InvalidUserOrGroupException,
+                                                         mock.ANY, mock.ANY,
+                                                         mock.ANY)
+            self.assertRaises(comm_meta.InvalidUserOrGroupException,
                               config.create_config_file)
 
     def test_destroy_monitored_metadata_proxy(self):
-        delete_if_exists = self.delete_if_exists.start()
-        mproxy_process = mock.Mock(
-            active=False, get_pid_file_name=mock.Mock(return_value='pid_file'))
+        mproxy_process = mock.Mock(active=False)
         mock_get_process = self.mock_get_process.start()
         mock_get_process.return_value = mproxy_process
         driver = metadata_driver.MetadataDriver(FakeL3NATAgent())
@@ -302,21 +349,19 @@ class TestMetadataDriverProcess(base.BaseTestCase):
                                                 'ns_name')
         mproxy_process.disable.assert_called_once_with(
             sig=str(int(signal.SIGTERM)))
-        delete_if_exists.assert_has_calls([
-            mock.call('pid_file', run_as_root=True)])
+        self.delete_if_exists.assert_called_once_with(
+            mock.ANY, run_as_root=True)
 
     def test_destroy_monitored_metadata_proxy_force(self):
-        delete_if_exists = self.delete_if_exists.start()
-        mproxy_process = mock.Mock(
-            active=True, get_pid_file_name=mock.Mock(return_value='pid_file'))
+        mproxy_process = mock.Mock(active=True)
         mock_get_process = self.mock_get_process.start()
         mock_get_process.return_value = mproxy_process
         driver = metadata_driver.MetadataDriver(FakeL3NATAgent())
-        with mock.patch.object(metadata_driver, 'SIGTERM_TIMEOUT', 0):
+        with mock.patch.object(driver_base, 'SIGTERM_TIMEOUT', 0):
             driver.destroy_monitored_metadata_proxy(mock.Mock(), 'uuid',
                                                     'conf', 'ns_name')
         mproxy_process.disable.assert_has_calls([
             mock.call(sig=str(int(signal.SIGTERM))),
             mock.call(sig=str(int(signal.SIGKILL)))])
-        delete_if_exists.assert_has_calls([
-            mock.call('pid_file', run_as_root=True)])
+        self.delete_if_exists.assert_called_once_with(
+            mock.ANY, run_as_root=True)
