@@ -13,6 +13,7 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+from concurrent import futures
 from unittest import mock
 
 from neutron_lib.api.definitions import port as port_def
@@ -77,7 +78,8 @@ class PortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
         mac_address = 'aa:aa:aa:aa:aa:aa'
         host_arg = {portbindings.HOST_ID: host,
                     'mac_address': mac_address}
-        with self.port(name='name', arg_list=(portbindings.HOST_ID,),
+        with self.port(name='name', is_admin=True,
+                       arg_list=(portbindings.HOST_ID,),
                        **host_arg) as port:
             self._check_response(port['port'], vif_type, has_port_filter,
                                  bound, status)
@@ -151,12 +153,12 @@ class PortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
             update_body = {'name': 'test_update'}
             if new_host is not None:
                 update_body[portbindings.HOST_ID] = new_host
-            with self.port(name='name', arg_list=(portbindings.HOST_ID,),
+            with self.port(name='name', is_admin=True,
+                           arg_list=(portbindings.HOST_ID,),
                            **host_arg) as port:
-                neutron_context = context.get_admin_context()
                 updated_port = self._update('ports', port['port']['id'],
                                             {'port': update_body},
-                                            neutron_context=neutron_context)
+                                            as_admin=True)
                 port_data = updated_port['port']
                 if new_host is not None:
                     self.assertEqual(new_host,
@@ -189,7 +191,7 @@ class PortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
         ctx = context.get_admin_context()
         plugin = directory.get_plugin()
         host_id = {portbindings.HOST_ID: 'host1'}
-        with self.port(**host_id) as port:
+        with self.port(is_admin=True, **host_id) as port:
             # Since the port is DOWN at first
             # It's necessary to make its status ACTIVE for this test
             plugin.update_port_status(ctx, port['port']['id'],
@@ -220,7 +222,8 @@ class PortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
 
     def test_distributed_binding(self):
         ctx = context.get_admin_context()
-        with self.port(device_owner=const.DEVICE_OWNER_DVR_INTERFACE) as port:
+        with self.port(is_admin=True,
+                       device_owner=const.DEVICE_OWNER_DVR_INTERFACE) as port:
             port_id = port['port']['id']
 
             # Verify port's VIF type and status.
@@ -234,7 +237,7 @@ class PortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
                  'device_id': 'router1'}})
 
             # Get port and verify VIF type and status unchanged.
-            port = self._show('ports', port_id)
+            port = self._show('ports', port_id, as_admin=True)
             self.assertEqual(portbindings.VIF_TYPE_DISTRIBUTED,
                              port['port'][portbindings.VIF_TYPE])
             self.assertEqual('DOWN', port['port']['status'])
@@ -246,7 +249,7 @@ class PortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
             self.assertEqual('local', details['network_type'])
 
             # Get port and verify VIF type and changed status.
-            port = self._show('ports', port_id)
+            port = self._show('ports', port_id, as_admin=True)
             self.assertEqual(portbindings.VIF_TYPE_DISTRIBUTED,
                              port['port'][portbindings.VIF_TYPE])
             self.assertEqual('BUILD', port['port']['status'])
@@ -257,7 +260,7 @@ class PortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
                 host='host-ovs-no_filter')
 
             # Get port and verify VIF type and changed status.
-            port = self._show('ports', port_id)
+            port = self._show('ports', port_id, as_admin=True)
             self.assertEqual(portbindings.VIF_TYPE_DISTRIBUTED,
                              port['port'][portbindings.VIF_TYPE])
             self.assertEqual('ACTIVE', port['port']['status'])
@@ -268,7 +271,7 @@ class PortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
                 host='host-ovs-no_filter')
 
             # Get port and verify VIF type and changed status.
-            port = self._show('ports', port_id)
+            port = self._show('ports', port_id, as_admin=True)
             self.assertEqual(portbindings.VIF_TYPE_DISTRIBUTED,
                              port['port'][portbindings.VIF_TYPE])
             self.assertEqual('DOWN', port['port']['status'])
@@ -351,6 +354,7 @@ class ExtendedPortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
         cfg.CONF.set_override('mechanism_drivers',
                               ['logger', 'test'],
                               'ml2')
+        cfg.CONF.set_override('extension_drivers', ['port_trusted'], 'ml2')
 
         driver_type.register_ml2_drivers_vlan_opts()
         cfg.CONF.set_override('network_vlan_ranges',
@@ -368,7 +372,10 @@ class ExtendedPortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
         if kwargs:
             data['binding'].update(kwargs)
         binding_resource = 'ports/%s/bindings' % port_id
-        binding_req = self.new_create_request(binding_resource, data, fmt)
+        binding_req = self.new_create_request(
+            binding_resource, data, fmt)
+        binding_req.environ['neutron.context'] = context.Context(
+            'service', 'service', roles=['service'])
         return binding_req.get_response(self.api)
 
     def _make_port_binding(self, fmt, port_id, host, **kwargs):
@@ -381,7 +388,8 @@ class ExtendedPortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
         data = {'binding': kwargs}
         binding_req = self.new_update_request('ports', data, port_id, fmt,
                                               subresource='bindings',
-                                              sub_id=host)
+                                              sub_id=host,
+                                              as_admin=True)
         return binding_req.get_response(self.api)
 
     def _do_update_port_binding(self, fmt, port_id, host, **kwargs):
@@ -391,10 +399,11 @@ class ExtendedPortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
         return self.deserialize(fmt, res)
 
     def _activate_port_binding(self, port_id, host, raw_response=True):
-        response = self._req('PUT', 'ports', id=port_id,
-                             data={'port_id': port_id},
-                             subresource='bindings', sub_id=host,
-                             action='activate').get_response(self.api)
+        response = self._service_req(
+            'PUT', 'ports', id=port_id,
+            data={'port_id': port_id},
+            subresource='bindings', sub_id=host,
+            action='activate').get_response(self.api)
         return self._check_code_and_serialize(response, raw_response)
 
     def _check_code_and_serialize(self, response, raw_response):
@@ -405,27 +414,33 @@ class ExtendedPortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
         return self.deserialize(self.fmt, response)
 
     def _list_port_bindings(self, port_id, params=None, raw_response=True):
-        response = self._req(
+        response = self._service_req(
             'GET', 'ports', fmt=self.fmt, id=port_id, subresource='bindings',
             params=params).get_response(self.api)
         return self._check_code_and_serialize(response, raw_response)
 
     def _show_port_binding(self, port_id, host, params=None,
                            raw_response=True):
-        response = self._req(
+        response = self._service_req(
             'GET', 'ports', fmt=self.fmt, id=port_id, subresource='bindings',
             sub_id=host, params=params).get_response(self.api)
         return self._check_code_and_serialize(response, raw_response)
 
     def _delete_port_binding(self, port_id, host):
-        response = self._req(
+        response = self._service_req(
             'DELETE', 'ports', fmt=self.fmt, id=port_id,
             subresource='bindings', sub_id=host).get_response(self.api)
         return response
 
-    def _create_port_and_binding(self, **kwargs):
-        device_owner = '%s%s' % (const.DEVICE_OWNER_COMPUTE_PREFIX, 'nova')
-        with self.port(device_owner=device_owner) as port:
+    # todo(slaweq): here I can add trusted to be checked too
+    def _create_port_and_binding(self, trusted=None, **kwargs):
+        port_kwargs = {
+            'device_owner': '%s%s' % (
+                const.DEVICE_OWNER_COMPUTE_PREFIX, 'nova')}
+        if trusted is not None:
+            port_kwargs['trusted'] = trusted
+            port_kwargs['is_admin'] = True
+        with self.port(**port_kwargs) as port:
             port_id = port['port']['id']
             binding = self._make_port_binding(self.fmt, port_id, self.host,
                                               **kwargs)['binding']
@@ -439,12 +454,13 @@ class ExtendedPortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
         self.assertEqual({'port_filter': False},
                          binding[pbe_ext.VIF_DETAILS])
 
-    def _assert_unbound_port_binding(self, binding):
+    def _assert_unbound_port_binding(self, binding, expected_profile=None):
         self.assertFalse(binding[pbe_ext.HOST])
         self.assertEqual(portbindings.VIF_TYPE_UNBOUND,
                          binding[pbe_ext.VIF_TYPE])
         self.assertEqual({}, binding[pbe_ext.VIF_DETAILS])
-        self.assertEqual({}, binding[pbe_ext.PROFILE])
+        expected_profile = expected_profile or {}
+        self.assertEqual(expected_profile, binding[pbe_ext.PROFILE])
 
     def test_create_port_binding(self):
         profile = {'key1': 'value1'}
@@ -456,7 +472,8 @@ class ExtendedPortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
     def test_create_duplicate_port_binding(self):
         device_owner = '%s%s' % (const.DEVICE_OWNER_COMPUTE_PREFIX, 'nova')
         host_arg = {portbindings.HOST_ID: self.host}
-        with self.port(device_owner=device_owner,
+        with self.port(is_admin=True,
+                       device_owner=device_owner,
                        arg_list=(portbindings.HOST_ID,),
                        **host_arg) as port:
             response = self._create_port_binding(self.fmt, port['port']['id'],
@@ -472,8 +489,7 @@ class ExtendedPortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
                                                  'host-fail')
             self.assertEqual(webob.exc.HTTPInternalServerError.code,
                              response.status_int)
-            self.assertTrue(exceptions.PortBindingError.__name__ in
-                            response.text)
+            self.assertIn(exceptions.PortBindingError.__name__, response.text)
 
     def test_create_port_binding_for_non_compute_owner(self):
         with self.port() as port:
@@ -529,8 +545,7 @@ class ExtendedPortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
                                                  self.host, **kwargs)
             self.assertEqual(webob.exc.HTTPInternalServerError.code,
                              response.status_int)
-            self.assertTrue(exceptions.PortBindingError.__name__ in
-                            response.text)
+            self.assertIn(exceptions.PortBindingError.__name__, response.text)
 
     def test_activate_port_binding(self):
         port, new_binding = self._create_port_and_binding()
@@ -539,7 +554,7 @@ class ExtendedPortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
             active_binding = self._activate_port_binding(
                 port['id'], self.host, raw_response=False)
         self._assert_bound_port_binding(active_binding)
-        updated_port = self._show('ports', port['id'])['port']
+        updated_port = self._show('ports', port['id'], as_admin=True)['port']
         updated_bound_drivers = updated_port[portbindings.VIF_DETAILS].pop(
             portbindings.VIF_DETAILS_BOUND_DRIVERS)
         self.assertEqual({'0': 'test'}, updated_bound_drivers)
@@ -562,6 +577,23 @@ class ExtendedPortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
             retrieved_bindings, const.INACTIVE)
         self._assert_unbound_port_binding(retrieved_inactive_binding)
 
+    def test_activate_port_binding_concurrency(self):
+        port, _ = self._create_port_and_binding()
+        with mock.patch.object(mechanism_test.TestMechanismDriver,
+                               '_check_port_context'):
+            with futures.ThreadPoolExecutor() as executor:
+                f1 = executor.submit(
+                    self._activate_port_binding, port['id'], self.host)
+                f2 = executor.submit(
+                    self._activate_port_binding, port['id'], self.host)
+                result_1 = f1.result()
+                result_2 = f2.result()
+
+        # One request should be successful and the other should receive a
+        # HTTPConflict. The order is arbitrary.
+        self.assertEqual({webob.exc.HTTPConflict.code, webob.exc.HTTPOk.code},
+                         {result_1.status_int, result_2.status_int})
+
     def test_activate_port_binding_for_non_compute_owner(self):
         port, new_binding = self._create_port_and_binding()
         data = {'port': {'device_owner': ''}}
@@ -587,8 +619,7 @@ class ExtendedPortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
             response = self._activate_port_binding(port['id'], self.host)
             self.assertEqual(webob.exc.HTTPInternalServerError.code,
                              response.status_int)
-            self.assertTrue(exceptions.PortBindingError.__name__ in
-                            response.text)
+            self.assertIn(exceptions.PortBindingError.__name__, response.text)
             self.assertEqual(ml2_plugin.MAX_BIND_TRIES, p_mock.call_count)
 
     def test_activate_port_binding_non_existing_binding(self):
@@ -618,12 +649,32 @@ class ExtendedPortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
         self.assertEqual(1, len(retrieved_bindings))
         self._assert_bound_port_binding(retrieved_bindings[0])
 
+    def test_list_port_bindings_with_trusted_field_set(self):
+        port, new_binding = self._create_port_and_binding(trusted=True)
+        retrieved_bindings = self._list_port_bindings(
+            port['id'], raw_response=False)['bindings']
+        self.assertEqual(2, len(retrieved_bindings))
+        self._assert_unbound_port_binding(
+            utils.get_port_binding_by_status_and_host(
+                retrieved_bindings, const.ACTIVE))
+        bound_port_binding = utils.get_port_binding_by_status_and_host(
+            retrieved_bindings, const.INACTIVE, host=self.host)
+        self._assert_bound_port_binding(bound_port_binding)
+        self.assertTrue(bound_port_binding['profile']['trusted'])
+
     def test_show_port_binding(self):
         port, new_binding = self._create_port_and_binding()
         retrieved_binding = self._show_port_binding(
             port['id'], self.host, raw_response=False)['binding']
         self._assert_bound_port_binding(retrieved_binding)
 
+    def test_show_port_binding_with_trusted_field_set(self):
+        port, new_binding = self._create_port_and_binding(trusted=True)
+        retrieved_binding = self._show_port_binding(
+            port['id'], self.host, raw_response=False)['binding']
+        self._assert_bound_port_binding(retrieved_binding)
+        self.assertTrue(retrieved_binding['profile']['trusted'])
+
     def test_show_port_binding_with_fields(self):
         port, new_binding = self._create_port_and_binding()
         fields = 'fields=%s' % pbe_ext.HOST
@@ -661,8 +712,7 @@ class ExtendedPortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
 
             self.assertEqual(webob.exc.HTTPInternalServerError.code,
                              response.status_int)
-            self.assertTrue(exceptions.PortBindingError.__name__ in
-                            response.text)
+            self.assertIn(exceptions.PortBindingError.__name__, response.text)
 
     def _create_unbound_port(self):
         with self.port() as port:
@@ -693,7 +743,8 @@ class ExtendedPortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
         with mock.patch.object(
                 mechanism_test.TestMechanismDriver, '_check_port_context'
         ):
-            req = self.new_update_request('ports', update_body, port_id)
+            req = self.new_update_request('ports', update_body, port_id,
+                                          as_service=True)
             self.assertEqual(200, req.get_response(self.api).status_int)
 
     def test_bind_non_pf_port_with_mac_port_not_updated(self):
@@ -833,7 +884,8 @@ class ExtendedPortBindingTestCase(test_plugin.NeutronDbPluginV2TestCase):
         with mock.patch.object(
                 mechanism_test.TestMechanismDriver, '_check_port_context'
         ):
-            req = self.new_update_request('ports', update_body, port['id'])
+            req = self.new_update_request('ports', update_body, port['id'],
+                                          as_service=True)
             self.assertEqual(200, req.get_response(self.api).status_int)
 
         # Neutron expected to reset the MAC to a generated one so that the
